from thinc.neural.util import get_array_module
from dataclasses import dataclass
import numpy

from .util import List, Array
from .util import lengths2mask


@dataclass
class RaggedArray:
    data: Array
    lengths: List[int]

    @property
    def size(self) -> int:
        return self.data.size

    @property
    def xp(self):
        return get_array_module(self.data)

    @property
    def dtype(self):
        return self.data.dtype

    @classmethod
    def blank(cls, xp=numpy) -> "RaggedArray":
        return RaggedArray(xp.zeros((0,), dtype="f"), [])

    @classmethod
    def from_truncated(cls, square: Array, lengths: List[int]) -> "RaggedArray":
        if len(lengths) != square.shape[0]:
            raise ValueError("Truncated array must have shape[0] == len(lengths)")
        width = square.shape[1]
        max_len = max(lengths, default=0)
        extra_dims = square.shape[2:]
        if width == max_len:
            return RaggedArray(square, lengths)
        elif width > max_len:
            raise ValueError("Expected width < max_len. Got {width} > {max_len}")
        xp = get_array_module(square)
        expanded = xp.zeros((sum(lengths),) + extra_dims, dtype=square.dtype)
        # TODO: I know there's a way to do this without the loop :(. Escapes
        # me currently.
        start = 0
        for i, length in enumerate(lengths):
            # We could have a row that's actually shorter than the width,
            # if the array was padded. Make sure we don't get junk values.
            row_width = min(width, length)
            expanded[start : start + row_width] = square[i, :row_width]
            start += length
        return cls(expanded, lengths)

    @classmethod
    def from_padded(cls, padded: Array, lengths: List[int]) -> "RaggedArray":
        if max(lengths, default=0) > padded.shape[1]:
            return cls.from_truncated(padded, lengths)
        mask = lengths2mask(lengths)
        assert sum(mask) == sum(lengths)
        all_rows = padded.reshape((-1,) + padded.shape[2:])
        xp = get_array_module(all_rows)
        data = xp.ascontiguousarray(all_rows[mask])
        assert data.shape[0] == sum(lengths)
        return cls(data, lengths)

    def to_padded(self, *, value=0, to: int = -1) -> Array:
        assert sum(self.lengths) == self.data.shape[0]
        max_len = max(self.lengths, default=0)
        if to >= 1 and to < max_len:
            raise ValueError(f"Cannot pad to {to}: Less than max length {max_len}")
        to = max(to, max_len)
        # Slightly convoluted implementation here, to do the operation in one
        # and avoid the loop
        shape = (len(self.lengths), to) + self.data.shape[1:]
        values = self.xp.zeros(shape, dtype=self.dtype)
        if self.data.size == 0:
            return values
        mask = lengths2mask(self.lengths)
        values = values.reshape((len(self.lengths) * to,) + self.data.shape[1:])
        values[mask] = self.data
        values = values.reshape(shape)
        return values

    def get(self, i: int) -> Array:
        start = sum(self.lengths[:i])
        end = start + self.lengths[i]
        return self.data[start:end]


@dataclass
class Activations:
    lh: RaggedArray
    po: RaggedArray

    @classmethod
    def blank(cls, *, xp=numpy):
        return cls(RaggedArray.blank(xp=xp), RaggedArray.blank(xp=xp))

    @property
    def xp(self):
        return self.lh.xp

    @property
    def has_lh(self) -> bool:
        return bool(self.lh.data.size)

    @property
    def has_po(self) -> bool:
        return bool(self.po.data.size)



import pytest
import numpy
from thinc.api import layerize
from spacy_pytorch_transformers.model_registry import with_length_batching
from spacy_pytorch_transformers.activations import Activations, RaggedArray


def make_activations(inputs, width):
    """Make an Activations object with the right sizing for an input."""
    lh = numpy.ones((inputs.data.shape[0], width), dtype="f")
    po = numpy.ones((len(inputs.lengths), width), dtype="f")
    return Activations(
        RaggedArray(lh, inputs.lengths), RaggedArray(po, [1 for _ in inputs.lengths])
    )


def create_dummy_model(width):
    record_fwd = []
    record_bwd = []

    @layerize
    def dummy_model(inputs, drop=0.0):
        def dummy_model_backward(d_outputs, sgd=None):
            record_bwd.append(d_outputs)
            return None

        record_fwd.append(inputs)
        return make_activations(inputs, width), dummy_model_backward

    return dummy_model, record_fwd, record_bwd


@pytest.mark.parametrize(
    "max_words,lengths,expected_batches",
    [(10, [4, 2, 2, 8, 2, 10], [[10], [8], [4], [2, 2, 2]]), (500, [9, 5], [[9, 5]])],
)
def test_with_length_batching(max_words, lengths, expected_batches, width=12):
    inputs = RaggedArray(numpy.arange(sum(lengths), dtype="i"), lengths)
    # The record_fwd and record_bwd variables will record the calls into the
    # forward and backward passes, respectively. We can use this to check that
    # the batching was done correctly.
    dummy, record_fwd, record_bwd = create_dummy_model(width)
    batched_dummy = with_length_batching(dummy, max_words)
    outputs, backprop = batched_dummy.begin_update(inputs)
    assert len(record_fwd) == len(expected_batches)
    assert [b.lengths for b in record_fwd] == expected_batches
    assert outputs.lh.data.shape == (inputs.data.shape[0], width)
    assert outputs.po.data.shape == (len(inputs.lengths), width)
    none = backprop(outputs)
    assert none is None
    assert len(record_bwd) == len(expected_batches)



import pytest
from numpy.testing import assert_equal
from spacy_pytorch_transformers import PyTT_Language, PyTT_WordPiecer
from spacy_pytorch_transformers import PyTT_TokenVectorEncoder, about
from spacy.attrs import LANG

from .util import make_tempdir, is_valid_tensor


def test_language_init(name):
    meta = {"lang": "en", "name": "test", "pipeline": []}
    nlp = PyTT_Language(meta=meta, pytt_name=name)
    assert nlp.lang == "en"
    assert nlp.meta["lang"] == "en"
    assert nlp.meta["lang_factory"] == PyTT_Language.lang_factory_name
    assert nlp.vocab.lang == "en"
    # Make sure we really have the EnglishDefaults here
    assert nlp.Defaults.lex_attr_getters[LANG](None) == "en"
    # Test requirements
    package = f"{about.__title__}>={about.__version__}"
    assert package in nlp.meta["requirements"]


def test_language_run(nlp):
    doc = nlp("hello world")
    assert is_valid_tensor(doc.tensor)


def test_language_wordpiece_to_from_bytes(name):
    nlp = PyTT_Language()
    nlp.add_pipe(nlp.create_pipe("sentencizer"))
    wordpiecer = PyTT_WordPiecer.from_pretrained(nlp.vocab, pytt_name=name)
    nlp.add_pipe(wordpiecer)
    doc = nlp("hello world")
    assert doc._.pytt_word_pieces is not None
    nlp2 = PyTT_Language()
    nlp2.add_pipe(nlp.create_pipe("sentencizer"))
    nlp2.add_pipe(PyTT_WordPiecer(nlp2.vocab))
    with pytest.raises(ValueError):
        new_doc = nlp2("hello world")
    nlp2.from_bytes(nlp.to_bytes())
    new_doc = nlp2("hello world")
    assert new_doc._.pytt_word_pieces is not None


def test_language_wordpiece_tok2vec_to_from_bytes(nlp, name):
    doc = nlp("hello world")
    assert is_valid_tensor(doc.tensor)
    nlp2 = PyTT_Language()
    nlp2.add_pipe(nlp2.create_pipe("sentencizer"))
    nlp2.add_pipe(PyTT_WordPiecer(nlp.vocab))
    nlp2.add_pipe(PyTT_TokenVectorEncoder(nlp.vocab))
    with pytest.raises(ValueError):
        new_doc = nlp2("hello world")
    nlp2.from_bytes(nlp.to_bytes())
    new_doc = nlp2("hello world")
    assert is_valid_tensor(new_doc.tensor)
    assert new_doc._.pytt_word_pieces is not None


def test_language_to_from_disk(nlp, name):
    doc = nlp("hello world")
    assert is_valid_tensor(doc.tensor)
    with make_tempdir() as tempdir:
        nlp.to_disk(tempdir)
        new_nlp = PyTT_Language()
        new_nlp.add_pipe(new_nlp.create_pipe("sentencizer"))
        wordpiecer = PyTT_WordPiecer(new_nlp.vocab, pytt_name=name)
        tok2vec = PyTT_TokenVectorEncoder(new_nlp.vocab, pytt_name=name)
        new_nlp.add_pipe(wordpiecer)
        new_nlp.add_pipe(tok2vec)
        new_nlp.from_disk(tempdir)
    assert new_nlp.pipe_names == nlp.pipe_names
    new_doc = new_nlp("hello world")
    assert is_valid_tensor(new_doc.tensor)
    assert_equal(doc.tensor, new_doc.tensor)



import pytest
from spacy_pytorch_transformers import PyTT_Language, PyTT_WordPiecer
from spacy_pytorch_transformers import PyTT_TokenVectorEncoder

MODEL_NAMES = ["bert-base-uncased", "gpt2", "xlnet-base-cased"]


@pytest.fixture(scope="session", params=MODEL_NAMES)
def name(request):
    return request.param


@pytest.fixture(scope="session")
def nlp(name):
    p_nlp = PyTT_Language(pytt_name=name)
    p_nlp.add_pipe(p_nlp.create_pipe("sentencizer"))
    p_nlp.add_pipe(PyTT_WordPiecer.from_pretrained(p_nlp.vocab, pytt_name=name))
    p_nlp.add_pipe(PyTT_TokenVectorEncoder.from_pretrained(p_nlp.vocab, name=name))
    return p_nlp



import pytest
from spacy_pytorch_transformers.pipeline import PyTT_TextCategorizer
from spacy.gold import GoldParse


@pytest.fixture(
    params=["softmax_last_hidden", "softmax_pooler_output", "softmax_class_vector"]
)
def textcat(name, nlp, request):
    arch = request.param
    width = nlp.get_pipe("pytt_tok2vec").model.nO
    textcat = PyTT_TextCategorizer(nlp.vocab, token_vector_width=width)
    textcat.add_label("Hello")
    config = {"architecture": arch, "pytt_name": name}
    if "gpt2" in name and arch in ("softmax_pooler_output", "softmax_class_vector"):
        with pytest.raises(ValueError):
            textcat.begin_training(**config)
        textcat.begin_training(pytt_name=name, architecture="softmax_last_hidden")
    elif "xlnet" in name and arch == "softmax_pooler_output":
        with pytest.raises(ValueError):
            textcat.begin_training(**config)
        textcat.begin_training(pytt_name=name, architecture="softmax_last_hidden")
    else:
        textcat.begin_training(**config)
    return textcat


def test_textcat_init(nlp):
    textcat = PyTT_TextCategorizer(nlp.vocab)
    assert textcat.labels == ()
    textcat.add_label("Hello")
    assert textcat.labels == ("Hello",)


def test_textcat_call(textcat, nlp):
    doc = nlp("hello world")
    for label in textcat.labels:
        assert label not in doc.cats
    doc = textcat(doc)
    for label in textcat.labels:
        assert label in doc.cats


def test_textcat_update(textcat, nlp):
    doc = nlp("hello world")
    optimizer = nlp.resume_training()
    cats = {"Hello": 1.0}
    losses = {}
    textcat.update([doc], [GoldParse(doc, cats=cats)], sgd=optimizer, losses=losses)
    assert "pytt_textcat" in losses


def test_textcat_update_multi_sentence(textcat, nlp):
    doc = nlp("Hello world. This is sentence 2.")
    assert len(list(doc.sents)) == 2
    optimizer = nlp.resume_training()
    cats = {"Hello": 1.0}
    losses = {}
    textcat.update([doc], [GoldParse(doc, cats=cats)], sgd=optimizer, losses=losses)
    assert "pytt_textcat" in losses


def test_textcat_update_batch(textcat, nlp):
    doc1 = nlp("Hello world. This is sentence 2.")
    doc2 = nlp("Hi again. This is sentence 4.")
    assert len(list(doc1.sents)) == 2
    assert len(list(doc2.sents)) == 2
    optimizer = nlp.resume_training()
    golds = [GoldParse(doc1, cats={"Hello": 1.0}), GoldParse(doc2, cats={"Hello": 0.0})]
    losses = {}
    textcat.update([doc1, doc2], golds, sgd=optimizer, losses=losses)
    assert "pytt_textcat" in losses



from configparser import ConfigParser


FIELDS = dict(
    max_seq_length=int,
    dropout=float,
    batch_size=int,
    eval_batch_size=int,
    learning_rate=float,
    weight_decay=float,
    adam_epsilon=float,
    max_grad_norm=float,
    num_train_epochs=int,
    max_steps=int,
    warmup_steps=int,
    seed=int,
    textcat_arch=str,
    eval_every=int,
    use_learn_rate_schedule=int,
    use_swa=int,
)


class HyperParams:
    pass


def get_hyper_params(path):
    cfg = ConfigParser()
    with path.open("r", encoding="utf8") as file_:
        cfg.read_string(file_.read())
    values = cfg["defaults"]
    HP = HyperParams()
    for field, type_ in FIELDS.items():
        setattr(HP, field, type_(values[field]))
    return HP



from typing import Union, List, Sequence, Callable, Any, Optional
import pytorch_transformers as pytt
import numpy
from spacy.tokens import Doc, Span

from . import _tokenizers

try:
    # This allows us to use cupy with mypy, for type checking
    import cupy  # noqa
except ImportError:
    pass

Array = Union["numpy.ndarray", "cupy.ndarray"]
Optimizer = Callable[[Array, Array, Optional[int]], None]
Dropout = Optional[float]


SPECIAL_TOKENS: Sequence[str] = (
    "[CLS]",
    "[BOS]",
    "[SEP]",
    "<cls>",
    "<sep>",
    "<|endoftext|>",
    "<s>",
    "</s>",
)


def get_pytt_config(name):
    """Map a name to the appropriate pytorch_transformers.*Config class."""
    name = name.lower()
    if "bert" in name:
        return pytt.BertConfig
    elif "xlnet" in name:
        return pytt.XLNetConfig
    elif "gpt2" in name:
        return pytt.GPT2Config
    elif "xlm" in name:
        return pytt.XLMConfig
    else:
        raise ValueError(f"Unsupported PyTT config name: '{name}'")


def get_pytt_model(name):
    """Map a name to the appropriate pytorch_transformers.*Model class."""
    name = name.lower()
    if "bert" in name:
        return pytt.BertModel
    elif "xlnet" in name:
        return pytt.XLNetModel
    elif "gpt2" in name:
        return pytt.GPT2Model
    elif "xlm" in name:
        return pytt.XLMModel
    else:
        raise ValueError(f"Unsupported PyTT config name: '{name}'")


def get_pytt_tokenizer(name):
    """Get a pytorch_transformers.*Tokenizer class from a name."""
    name = name.lower()
    if "bert" in name:
        return _tokenizers.SerializableBertTokenizer
    elif "xlnet" in name:
        return _tokenizers.SerializableXLNetTokenizer
    elif "gpt2" in name:
        return _tokenizers.SerializableGPT2Tokenizer
    elif "xlm" in name:
        return _tokenizers.SerializableXLMTokenizer
    else:
        raise ValueError(f"Unsupported PyTT config name: '{name}'")


def pad_batch(
    batch: List[Array], *, axis: int = 0, xp=numpy, to: int = 0, value: int = -1
) -> Array:
    """Pad a batch of arrays with zeros so that sequences are the same
    length, and form them into a single array.
    """
    if not batch:
        return xp.zeros((0, to))

    if isinstance(batch[0], list):
        batch = [xp.array(x, dtype=numpy.int_) for x in batch]

    max_len = max((seq.shape[axis] for seq in batch), default=0)
    if to < 1:
        to = max_len
    elif max_len > to:
        raise ValueError(f"Cannot pad_batch with max len {max_len} to {to}.")

    if isinstance(batch[0], list) or len(batch[0].shape) == 1:
        return _pad_batch_1d(batch, xp=xp, to=to, value=value)
    else:
        return _pad_batch_nd(batch, axis=axis, xp=xp, to=to, value=value)


def _pad_batch_1d(batch: List[Array], *, xp=numpy, to: int, value) -> Array:
    """Pad a batch of lists or 1d arrays with zeros so that sequences are the same
    length, and form them into a single array.
    """
    padded: List[Array] = []
    seq: Array
    values = (0, value)
    pad_desc = [[0, 0]]
    for seq in batch:
        pad_desc[0][1] = to - len(seq)
        padded.append(xp.pad(seq, pad_desc, mode="constant", constant_values=values))
    output = xp.vstack(padded)
    assert output.shape == (len(batch), to), output.shape
    return output


def _pad_batch_nd(
    batch: List[Array], axis: int, *, xp=numpy, to: int = 0, value=-1
) -> Array:
    padded: List[Array] = []
    seq: Array
    values = (0, value)
    pad_desc = [[0, 0] for _ in batch[0].shape]
    for seq in batch:
        # Ugh, numpy.pad sucks.
        pad_desc[axis][1] = to - seq.shape[axis]
        arr = xp.pad(seq, pad_desc, mode="constant", constant_values=values)
        if len(arr.shape) == 2:
            # This prevents us concatenating on the sequence dimension, when what
            # we want is to have a new batch dimension.
            arr = arr.reshape((1, arr.shape[0], arr.shape[1]))
        padded.append(arr)
    output = xp.vstack(padded)
    return output


def batch_by_length(seqs: Union[List[Array]], max_words: int) -> List[List[int]]:
    """Given a list of sequences, return a batched list of indices into the
    list, where the batches are grouped by length, in descending order. Batches
    may be at most max_words in size, defined as max sequence length * size.
    """
    # Use negative index so we can get sort by position ascending.
    lengths_indices = [(len(seq), i) for i, seq in enumerate(seqs)]
    lengths_indices.sort()
    batches: List[List[int]] = []
    batch: List[int] = []
    for length, i in lengths_indices:
        # i = -neg_i
        if not batch:
            batch.append(i)
        elif length * (len(batch) + 1) <= max_words:
            batch.append(i)
        else:
            batches.append(batch)
            batch = [i]
    if batch:
        batches.append(batch)
    # Check lengths match
    assert sum(len(b) for b in batches) == len(seqs)
    # Check no duplicates
    seen = set()
    for b in batches:
        seen.update(b)
    assert len(seen) == len(seqs)
    batches = [list(sorted(batch)) for batch in batches]
    batches.reverse()
    return batches


def ensure3d(arr: Array, *, axis: int = 1) -> Array:
    """Make sure an array is 3d, inserting a dimension at axis if not."""
    if arr.size == 0:
        return arr.reshape((0, 0, 0))
    elif len(arr.shape) == 3:
        return arr
    elif len(arr.shape) == 2:
        return arr.reshape((arr.shape[0], 1, arr.shape[1]))
    else:
        raise ValueError(f"Cannot make array 3d. Shape: {arr.shape}")


def unflatten_list(flat: List[Any], lengths: List[int]) -> List[List[Any]]:
    """Unflatten a list into nested sublists, where each sublist i should have
    length lengths[i]."""
    nested: List[List[Any]] = []
    offset = 0
    for length in lengths:
        nested.append(flat[offset : offset + length])
        offset += length
    return nested


def flatten_list(nested: List[List[Any]]) -> List[Any]:
    """Flatten a nested list."""
    flat = []
    for x in nested:
        flat.extend(x)
    return flat


def lengths2mask(lengths):
    """Get a boolean mask of which entries in a padded batch are valid, given
    a list of lengths."""
    padded = pad_batch([numpy.ones((L,), dtype="i") for L in lengths])
    padded[padded < 0] = 0
    return padded.reshape((-1,)) >= 1


def is_special_token(text: str) -> bool:
    return text in SPECIAL_TOKENS


def is_class_token(text: str) -> bool:
    return text == "[CLS]" or text == "<cls>"


def get_segment_ids(name: str, *lengths) -> List[int]:
    if len(lengths) == 1:
        length1 = lengths[0]
        length2 = 0
    elif len(lengths) == 2:
        length1, length2 = lengths
    else:
        msg = f"Expected 1 or 2 segments. Got {len(lengths)}"
        raise ValueError(msg)
    if "bert" in name:
        return get_bert_segment_ids(length1, length2)
    elif "xlnet" in name:
        return get_xlnet_segment_ids(length1, length2)
    elif "xlm" in name:
        return get_xlm_segment_ids(length1, length2)
    elif "gpt2" in name:
        return get_gpt2_segment_ids(length1, length2)
    else:
        raise ValueError(f"Unexpected model name: {name}")


def get_bert_segment_ids(length1: int, length2: int) -> List[int]:
    """Get an array of segment IDs in BERT's format, for an input with one or
    two segments (set length2=0 for one segment). The lengths should be just the
    wordpiece lengths, not including the SEP and CLS tokens.

    According to the HF glue_utils.py module, the convention for BERT is:

    (a) For sequence pairs:
        tokens: [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        type_ids:   0   0  0    0    0     0     0   0   1  1  1  1   1   1

    (b) For single sequences:
        tokens:   [CLS] the dog is hairy . [SEP]
        type_ids:   0   0   0   0  0     0   0
    """
    if length2:
        return [0] * length1 + [0] + [0] + [1] * length2 + [1]
    else:
        return [0] * length1 + [0] + [0]


def get_xlnet_segment_ids(length1: int, length2: int) -> List[int]:
    """Get an array of segment IDs in XLNet's format, for an input with one or
    two segments (set length2=0 for one segment). The lengths should be just the
    wordpiece lengths, not including the SEP and CLS tokens.

    According to the XLNet code classifer_utils.py module, the convention is:

    (a) For sequence pairs:
        tokens:    is this jack ##son ##ville ? <sep> no it is not . <sep> <cls>
        type_ids:   0   0  0    0      0      0  0    1   1  1  1  1   1   2

    (b) For single sequences:
        tokens:   the dog is hairy . <sep> <cls>
        type_ids:   0   0 0   0    0   0   2
    """
    if length2:
        return [0] * length1 + [0] + [1] * length2 + [1, 2]
    else:
        return [0] * length1 + [0, 2]


def get_xlm_segment_ids(length1: int, length2: int) -> List[int]:
    """Get an array of segment IDs in XLNet's format, for an input with one or
    two segments (set length2=0 for one segment). The lengths should be just the
    wordpiece lengths, not including the SEP and CLS tokens.

    According to the HF glue_utils.py script, the convention is:

    (a) For sequence pairs:
        tokens: [CLS] is this jack ##son ##ville ? [SEP] no it is not . [SEP]
        type_ids:   0   0  0    0    0     0     0   0   1  1  1  1   1   1

    (b) For single sequences:
        tokens:   [CLS] the dog is hairy . [SEP]
        type_ids:   0   0   0   0  0     0   0
    """
    if length2:
        return [0] * length1 + [0] + [0] + [1] * length2 + [1]
    else:
        [0] * length1 + [0] + [0]


def get_gpt2_segment_ids(length1: int, length2: int) -> List[int]:
    """Get an array of segment IDs in GPT2's format, for an input with one or
    two segments (set length2=0 for one segment). The lengths should be just the
    wordpiece lengths, not including the SEP and CLS tokens.

    I'm really not sure how this should look? We currently require segment
    boundaries, so we're just using the <|endoftext|> markers in their vocab?

    (a) For sequence pairs:
        tokens:   <|eot|> is this jack ##son ##ville ? <|eot|> no it is not . <|eot|>
        type_ids:   0     0  0    0    0     0       0  0      1  1  1  1   1 1

    (b) For single sequences:
        tokens:   <|eot|> the dog is hairy . <|eot|>
        type_ids:   0      0   0   0   0   0 0
    """
    if not length2:
        return [0] + [0] * length1 + [0]
    else:
        return [0] + [0] * length1 + [0] + [1] * length2 + [1]


def get_sents(doc: Union[Span, Doc]) -> List[Span]:
    if doc.is_sentenced:
        return list(doc.sents)
    else:
        return [doc[:]]


def warmup_linear_rates(initial_rate, warmup_steps, total_steps):
    """Generate a series, starting from an initial rate, and then with a warmup
    period, and then a linear decline. Used for learning rates.
    """
    step = 0
    while True:
        if step < warmup_steps:
            factor = step / max(1, warmup_steps)
        else:
            factor = max(
                0.0, (total_steps - step) / max(1.0, total_steps - warmup_steps)
            )
        yield factor * initial_rate
        step += 1


from .activations import Activations  # noqa



from spacy.language import Language
from spacy.tokens import Doc, Span, Token
from spacy.util import get_lang_class
from spacy.gold import GoldParse

from .util import is_special_token
from . import about


class PyTT_Language(Language):
    """A subclass of spacy.Language that holds a PyTorch-Transformer (PyTT) pipeline.

    PyTT pipelines work only slightly differently from spaCy's default pipelines.
    Specifically, we introduce a new pipeline component at the start of the pipeline,
    PyTT_TokenVectorEncoder. We then modify the nlp.update() function to run
    the PyTT_TokenVectorEncoder before the other pipeline components, and
    backprop it after the other components are done.
    """

    lang_factory_name = "pytt"

    @staticmethod
    def install_extensions():
        tok2vec_attrs = [
            "pytt_last_hidden_state",
            "pytt_pooler_output",
            "pytt_all_hidden_states",
            "pytt_all_attentions",
            "pytt_d_last_hidden_state",
            "pytt_d_pooler_output",
            "pytt_d_all_hidden_states",
            "pytt_d_all_attentions",
        ]
        for attr in tok2vec_attrs:
            Doc.set_extension(attr, default=None)
            Span.set_extension(attr, getter=get_span_tok2vec_getter(attr))
            Token.set_extension(attr, getter=get_token_tok2vec_getter(attr))
        wp_attrs = ["pytt_alignment", "pytt_word_pieces", "pytt_word_pieces_"]
        for attr in wp_attrs:
            Doc.set_extension(attr, default=None)
            Span.set_extension(attr, getter=get_span_wp_getter(attr))
            Token.set_extension(attr, getter=get_token_wp_getter(attr))
        Doc.set_extension("pytt_separator", default=None)
        Span.set_extension(
            "pytt_separator", getter=lambda span: span.doc._.pytt_separator
        )
        Token.set_extension(
            "pytt_separator", getter=lambda token: token.doc._.pytt_separator
        )
        Doc.set_extension("pytt_segments", getter=get_pytt_segments)
        Span.set_extension("pytt_segments", getter=get_pytt_segments)
        for cls in [Token, Span, Doc]:
            cls.set_extension("pytt_start", getter=get_wp_start)
            cls.set_extension("pytt_end", getter=get_wp_end)

    def __init__(
        self, vocab=True, make_doc=True, max_length=10 ** 6, meta={}, **kwargs
    ):
        """Initialize the language class. Expects either a pytt_name setting in
        the meta or as a keyword argument, specifying the pre-trained model
        name. This is used to set up the model-specific tokenizer.
        """
        meta = dict(meta)
        meta["lang_factory"] = self.lang_factory_name
        # Add this package to requirements to it will be included in the
        # install_requires of any model using this language class
        package = f"{about.__title__}>={about.__version__}"
        meta.setdefault("requirements", []).append(package)
        self.lang = meta.get("lang", "xx")
        self.Defaults = get_defaults(self.lang)
        super().__init__(vocab, make_doc, max_length, meta=meta, **kwargs)

    def update(self, docs, golds, drop=0.0, sgd=None, losses=None, component_cfg={}):
        component_cfg = dict(component_cfg)
        if self.has_pipe("sentencizer"):
            sentencizer = self.get_pipe("sentencizer")
        else:
            sentencizer = lambda doc: doc
        if self.has_pipe("pytt_wordpiecer"):
            wp = self.get_pipe("pytt_wordpiecer")
        else:
            wp = lambda doc: doc
        tok2vec = self.get_pipe("pytt_tok2vec")
        new_docs = []
        new_golds = []
        for doc, gold in zip(docs, golds):
            if isinstance(doc, str):
                doc = self.make_doc(doc)
                doc = sentencizer(doc)
                doc = wp(doc)
            if not isinstance(gold, GoldParse):
                gold = GoldParse(doc, **gold)
            new_docs.append(doc)
            new_golds.append(gold)
        docs = new_docs
        golds = new_golds
        pytt_outputs, backprop_tok2vec = tok2vec.begin_update(
            docs, drop=drop, **component_cfg.get("pytt_tok2vec", {})
        )
        tok2vec.set_annotations(docs, pytt_outputs)
        for doc in docs:
            assert doc._.pytt_last_hidden_state is not None
        with self.disable_pipes("pytt_tok2vec"):
            super().update(
                docs,
                golds,
                drop=0.1,
                sgd=sgd,
                losses=losses,
                component_cfg=component_cfg,
            )
        backprop_tok2vec(docs, sgd=sgd)

    def resume_training(self, sgd=None, component_cfg=None, **kwargs):
        """Continue training a pre-trained model.

        Before running the normal Language.resume_training method, we do the
        following:

        * Look for a tok2vec pipeline component. By default we look for the name
            'pytt_tok2vec'. This can be changed with the tok2vec_name keyword
            argument. If no component is found, a ValueError is raised.
        * If any other components have `component.model == True` and a `.begin_training()`
            method, we call the `.begin_training()` method. Configuration can
            be passed in using the component_cfg keyword argument. If unset,
            we also pass in a value for token_vector_width, which we read from
            the tok2vec component.
        """
        if component_cfg is None:
            component_cfg = {}
        tok2vec_name = kwargs.get("tok2vec_name", "pytt_tok2vec")
        tok2vec = self.get_pipe(tok2vec_name)
        token_vector_width = tok2vec.token_vector_width
        for name, component in self.pipeline:
            if name == tok2vec_name:
                continue
            elif getattr(component, "model", None) is not True:
                continue
            elif not hasattr(component, "begin_training"):
                continue
            cfg = component_cfg.get(name, {})
            if "tok2vec_name" not in component_cfg:
                cfg["tok2vec_name"] = tok2vec_name
            if "token_vector_width" not in component_cfg:
                cfg["token_vector_width"] = token_vector_width
            component.cfg.update(cfg)
            component.begin_training(pipeline=self.pipeline, sgd=False, **cfg)
            assert component.model is not True
        optimizer = super().resume_training(sgd=sgd, **kwargs)
        optimizer.L2 = 0.0
        return optimizer


def get_defaults(lang):
    """Get the language-specific defaults, if available in spaCy."""
    try:
        lang_cls = get_lang_class(lang)
        return lang_cls.Defaults
    except ImportError:
        return Language.Defaults


def get_wp_start(span):
    if isinstance(span, Token):
        span = span.doc[span.i : span.i + 1]
    for token in span:
        if token._.pytt_alignment:
            wp_start = token._.pytt_alignment[0]
            break
    else:
        return None
    wordpieces = span.doc._.pytt_word_pieces_
    # This is a messy way to check for the XLNet-style pattern, where we can
    # have <sep> <cls>. In the BERT-style pattern, we have [cls] at start.
    if is_special_token(wordpieces[0]):
        if wp_start >= 1 and is_special_token(wordpieces[wp_start - 1]):
            return wp_start - 1
    return wp_start


def get_wp_end(span):
    if isinstance(span, Token):
        span = span.doc[span.i : span.i + 1]
    for token in reversed(span):
        if token._.pytt_alignment:
            wp_end = token._.pytt_alignment[-1]
            break
    else:
        return None
    wordpieces = span.doc._.pytt_word_pieces_
    if (wp_end + 1) < len(wordpieces) and is_special_token(wordpieces[wp_end + 1]):
        wp_end += 1
    # This is a messy way to check for the XLNet-style pattern, where we can
    # have <sep> <cls>. In the BERT-style pattern, we have [cls] at start.
    if not is_special_token(wordpieces[0]):
        if (wp_end + 1) < len(wordpieces) and is_special_token(wordpieces[wp_end + 1]):
            wp_end += 1
    return wp_end


def get_span_wp_getter(attr):
    def span_alignment_getter(span):
        return [token._.get(attr) for token in span]

    def span_getter(span):
        start = span._.pytt_start
        end = span._.pytt_end
        if start is None and end is None:
            return []
        doc_values = span.doc._.get(attr)
        start = start if start is not None else 0
        if end is None:
            return doc_values[start:]
        return doc_values[start : end + 1]

    if attr == "pytt_alignment":
        return span_alignment_getter
    else:
        return span_getter


def get_token_wp_getter(attr):
    def token_alignment_getter(token):
        doc_values = token.doc._.get(attr)
        return doc_values[token.i] if doc_values is not None else None

    def token_wordpiece_getter(token):
        doc_values = token.doc._.get(attr)
        start = token._.pytt_start
        end = token._.pytt_end
        if start is None and end is None:
            return []
        return [doc_values[i] for i in range(start, end + 1)]

    if attr == "pytt_alignment":
        return token_alignment_getter
    else:
        return token_wordpiece_getter


def get_span_tok2vec_getter(attr):
    def span_getter(span):
        doc_activations = span.doc._.get(attr)
        if doc_activations is None:
            return None
        wp_start = span[0]._.pytt_wp_start
        wp_end = span[-1]._.pytt_wp_end
        if wp_start is not None and wp_end is not None:
            return doc_activations[wp_start : wp_end + 1]
        else:
            # Return empty slice.
            return doc_activations[0:0]

    return span_getter


def get_token_tok2vec_getter(attr):
    def token_getter(token):
        # Delegate through span, so get a span with just the token.
        span = token.doc[token.i : token.i + 1]
        return span._.get(attr)

    return token_getter


def get_pytt_segments(doc):
    separator = doc._.pytt_separator
    if separator is not None:
        start = 0
        for token in doc:
            if token.text == separator:
                yield doc[start : token.i + 1]
                start = token.i + 1
        yield doc[start:]
    else:
        yield doc[:]



from thinc.extra.wrappers import PyTorchWrapper, xp2torch, torch2xp
from pytorch_transformers.optimization import AdamW
import pytorch_transformers as pytt
import torch.autograd
import torch.nn.utils.clip_grad
import torch
from typing import Tuple, Callable, Any
from thinc.neural.optimizers import Optimizer
import numpy
from torchcontrib.optim import SWA

from .util import get_pytt_model
from .util import Dropout
from .activations import RaggedArray, Activations

FINE_TUNE = True
CONFIG = {"output_hidden_states": True, "output_attentions": True}


class PyTT_Wrapper(PyTorchWrapper):
    """Wrap a PyTorch-Transformers model for use in Thinc.

    The model will take as input a spacy_pytorch_transformers.util.RaggedArray
    object that will specify the input IDs and optionally the segment IDs. The
    RaggedArray is basically a tuple (ids, lengths), where ids is concatenated
    for a whole batch (this format allows the data to be contiguous even if
    the sequences are different lengths). The segment IDs should be coded as
    the different models expect them -- see
    https://github.com/huggingface/pytorch-transformers/blob/master/examples/utils_glue.py
    """

    _model: Any
    _optimizer: Any
    cfg: dict

    @classmethod
    def from_pretrained(cls, name):
        model_cls = get_pytt_model(name)
        model = model_cls.from_pretrained(name, **CONFIG)
        self = cls(name, model.config.to_dict(), model)
        self.cfg.update(self.pytt_model.config.to_dict())
        return self

    def __init__(self, name, config, model):
        PyTorchWrapper.__init__(self, model)
        self.cfg = dict(config)

    @property
    def nO(self):
        if "hidden_size" in self.cfg:
            # BERT
            return self.cfg["hidden_size"]
        elif "n_embd" in self.cfg:
            # GPT2
            return self.cfg["n_embd"]
        elif "d_model" in self.cfg:
            # XLNet
            return self.cfg["d_model"]
        elif hasattr(self.pytt_model, "dim"):
            # XLM
            return self.pytt_model.dim
        else:
            keys = ", ".join(self.cfg.keys())
            raise ValueError(f"Unexpected config. Keys: {keys}")

    @property
    def pytt_model(self):
        return self._model

    @property
    def max_length(self):
        return self.cfg.get("max_position_embeddings", 128)

    def predict(self, inputs: RaggedArray):
        self._model.eval()
        model_kwargs = self.get_model_kwargs(inputs)
        with torch.no_grad():
            if hasattr(self._optimizer, "swap_swa_sgd"):
                self._optimizer.swap_swa_sgd()
            y_var = self._model(**model_kwargs)
            if hasattr(self._optimizer, "swap_swa_sgd"):
                self._optimizer.swap_swa_sgd()
        return self.make_activations(y_var, inputs.lengths)

    def begin_update(
        self, inputs: RaggedArray, drop: Dropout = 0.0
    ) -> Tuple[Activations, Callable[..., None]]:
        if drop is None:
            # "drop is None" indicates prediction. It's one of the parts of
            # Thinc's API I'm least happy with...
            return self.predict(inputs), lambda dY, sgd=None: None
        max_original = max(inputs.lengths, default=0)
        model_kwargs = self.get_model_kwargs(inputs)
        self._model.train()
        # Prepare all the model arguments, including the attention mask
        y_var = self._model(**model_kwargs)
        output = self.make_activations(y_var, inputs.lengths)
        assert output.lh.data.shape[0] == inputs.data.shape[0], (
            output.lh.data.shape,
            inputs.data.shape,
        )

        def backward_pytorch(d_output: Activations, sgd: Optimizer = None) -> None:
            y_for_bwd = []
            dy_for_bwd = []
            if d_output.has_lh:
                assert d_output.lh.data.shape[0] == sum(d_output.lh.lengths)
                d_lh = d_output.lh.to_padded(to=max_original)
                if self.max_length and d_lh.shape[1] >= self.max_length:
                    d_lh = d_lh[:, : self.max_length]
                dy_for_bwd.append(xp2torch(d_lh))
                y_for_bwd.append(y_var[0])
            if d_output.has_po:
                dy_for_bwd.append(xp2torch(d_output.po.data))
                y_for_bwd.append(y_var[1])
            if FINE_TUNE:
                torch.autograd.backward(y_for_bwd, grad_tensors=dy_for_bwd)
                if sgd is not None:
                    if self._optimizer is None:
                        self._optimizer = self._create_optimizer(sgd)
                    if sgd.max_grad_norm:
                        torch.nn.utils.clip_grad.clip_grad_norm_(
                            self._model.parameters(), sgd.max_grad_norm
                        )
                    optimizer = self._optimizer

                    for params in optimizer.param_groups:
                        params["lr"] = getattr(sgd, "pytt_lr", sgd.alpha)
                    optimizer.step()
                    optimizer.zero_grad()
            return None

        self._model.eval()
        return output, backward_pytorch

    def make_activations(self, fields, lengths) -> Activations:
        """Create Activations from the output tuples produced by PyTorch Transformers.
        Includes converting torch tensors to xp, and handling missing values.
        """
        fields = list(fields)
        fields[0] = torch2xp(fields[0])
        fields[0] = RaggedArray.from_padded(fields[0], lengths)
        assert fields[0].data.shape[0] == sum(lengths)
        # lh: last hidden
        # po: pooler_output
        # ah: all_hidden
        # aa: all_attention
        if len(fields) != 4:
            lh = fields[0]
            po = RaggedArray.blank()
        else:
            if isinstance(fields[1], tuple):
                fields[1] = RaggedArray.blank()
            else:
                fields[1] = RaggedArray(torch2xp(fields[1]), [1] * len(lengths))
            lh, po, _, _2 = fields
        # Convert last_hidden_state to xp
        return Activations(lh, po)

    def get_model_kwargs(self, inputs):
        padded = inputs.to_padded()
        if padded.ndim == 2:
            padded = padded.reshape(padded.shape + (1,))
        if self.max_length:
            padded = padded[:, : self.max_length]
        ids = padded[:, :, 0]
        neg_idx = ids < 0
        ids[neg_idx] = 0
        ids = torch.as_tensor(ids, dtype=torch.int64)
        if padded.shape[2] == 2:
            segment_ids = padded[:, :, 1]
            segment_ids = torch.as_tensor(segment_ids, dtype=torch.int64)
        else:
            segment_ids = torch.zeros_like(ids)
        # Calculate "attention mask" for BERT and  XLNet, but not GPT2 (sigh)
        if isinstance(self._model, (pytt.BertModel, pytt.XLNetModel)):
            mask = self.ops.xp.ones(ids.shape, dtype=numpy.int_)
            mask[neg_idx] = 0
            mask = xp2torch(mask)
            return {
                "input_ids": ids,
                "attention_mask": mask,
                "token_type_ids": segment_ids,
            }
        else:
            return {"input_ids": ids, "token_type_ids": segment_ids}

    def _create_optimizer(self, sgd):
        optimizer = AdamW(
            self._model.parameters(),
            lr=getattr(sgd, "pytt_lr", sgd.alpha),
            eps=sgd.eps,
            betas=(sgd.b1, sgd.b2),
            weight_decay=getattr(sgd, "pytt_weight_decay", 0.0),
        )
        if getattr(sgd, "pytt_use_swa", False):
            lr = getattr(sgd, "pytt_lr", sgd.alpha)
            optimizer = SWA(optimizer, swa_start=1, swa_freq=1, swa_lr=lr)
        optimizer.zero_grad()
        return optimizer



__title__ = "spacy-pytorch-transformers"
__version__ = "0.2.0"
__summary__ = "spaCy pipelines for pre-trained BERT and other transformers"
__uri__ = "https://explosion.ai"
__author__ = "Explosion"
__email__ = "contact@explosion.ai"
__license__ = "MIT"



import spacy.pipeline
from ..model_registry import get_model_function

DEBUG_LOSS = False


class PyTT_TextCategorizer(spacy.pipeline.TextCategorizer):
    """Subclass of spaCy's built-in TextCategorizer component that supports
    using the features assigned by the PyTorch-Transformers models via the token
    vector encoder. It requires the PyTT_TokenVectorEncoder to run before it in
    the pipeline.
    """

    name = "pytt_textcat"

    @classmethod
    def from_nlp(cls, nlp, **cfg):
        """Factory to add to Language.factories via entry point."""
        return cls(nlp.vocab, **cfg)

    @classmethod
    def Model(cls, nr_class=1, exclusive_classes=False, **cfg):
        """Create a text classification model using a PyTorch-Transformers model
        for token vector encoding.

        nr_class (int): Number of classes.
        width (int): The width of the tensors being assigned.
        exclusive_classes (bool): Make categories mutually exclusive.
        **cfg: Optional config parameters.
        RETURNS (thinc.neural.Model): The model.
        """
        arch = cfg.get("architecture", "softmax_class_vector")
        # This is optional -- but if it's set, we can debug config errors.
        pytt_name = cfg.get("pytt_name", "")
        is_gpt2 = "gpt2" in pytt_name
        is_xlnet = "xlnet" in pytt_name
        msg = (
            f"PyTT_TextCategorizer model architecture set to '{arch}' "
            f"with {pytt_name} transformer. This "
            f"combination is incompatible, as the transformer does not "
            f"provide that output feature."
        )
        if is_gpt2 and arch in ("softmax_class_vector", "softmax_pooler_output"):
            raise ValueError(msg)
        elif is_xlnet and arch == "softmax_pooler_output":
            raise ValueError(msg)
        make_model = get_model_function(arch)
        return make_model(nr_class, exclusive_classes=exclusive_classes, **cfg)

    def get_loss(self, docs, golds, scores):
        # This is a useful diagnostic while figuring out whether your model is
        # learning anything. We print the loss each batch, and also the
        # mean and variance for the score of the first class.
        # If the model is learning things, we want to see the mean score stay
        # close the the class distribution (e.g. 0.5 for balanced classes),
        # while the variance in scores should increase (i.e. different examples
        # should get different scores).
        loss, d_scores = super().get_loss(docs, golds, scores)
        mean_score = scores.mean(axis=0)[0]
        var_score = scores.var(axis=0)[0]
        if DEBUG_LOSS:
            print("L", "%.4f" % loss, "m", "%.3f" % mean_score, "v", "%.6f" % var_score)
        return loss, d_scores

    def begin_training(
        self, get_gold_tuples=lambda: [], pipeline=None, sgd=None, **kwargs
    ):
        if self.model is True:
            self.cfg.update(kwargs)
            self.require_labels()
            self.model = self.Model(len(self.labels), **self.cfg)
        if sgd is None:
            sgd = self.create_optimizer()
        return sgd



from .language import PyTT_Language
from .pipeline.tok2vec import PyTT_TokenVectorEncoder  # noqa
from .pipeline.textcat import PyTT_TextCategorizer  # noqa
from .pipeline.wordpiecer import PyTT_WordPiecer  # noqa
from .model_registry import register_model, get_model_function  # noqa
from .about import __version__  # noqa

PyTT_Language.install_extensions()



import pytest
from spacy_pytorch_transformers import PyTT_WordPiecer
from spacy.tokens import Doc
from spacy.vocab import Vocab


@pytest.fixture
def vocab():
    return Vocab()


@pytest.fixture
def wordpiecer(name):
    return PyTT_WordPiecer.from_pretrained(vocab, pytt_name=name)


def test_alignment_extension_attr(vocab):
    doc = Doc(vocab, words=["hello", "world", "test"])
    doc._.pytt_alignment = [[1, 2], [3, 4], [5, 6]]
    assert doc[0]._.pytt_alignment == [1, 2]
    assert doc[1]._.pytt_alignment == [3, 4]
    assert doc[2]._.pytt_alignment == [5, 6]
    assert doc[0:2]._.pytt_alignment == [[1, 2], [3, 4]]
    assert doc[1:3]._.pytt_alignment == [[3, 4], [5, 6]]


def test_wp_span_extension_attr(name, vocab, wordpiecer):
    if name == "gpt2":
        return
    doc = Doc(vocab, words=["hello", "world"])
    for w in doc[1:]:
        w.is_sent_start = False
    doc = wordpiecer(doc)
    assert len(doc._.pytt_word_pieces) == 4
    assert doc[0]._.pytt_start == 0
    assert doc[-1]._.pytt_end == 3
    assert len(doc[:]._.pytt_word_pieces) == 4



import pytest
from spacy_pytorch_transformers import PyTT_WordPiecer
from spacy_pytorch_transformers.util import is_special_token, get_pytt_tokenizer
from spacy.vocab import Vocab
from spacy.tokens import Doc


@pytest.fixture(scope="session")
def wp(name):
    return PyTT_WordPiecer.from_pretrained(Vocab(), pytt_name=name)


def test_wordpiecer(wp):
    words = ["hello", "world", "this", "is", "a", "test"]
    doc = Doc(wp.vocab, words=words)
    doc[0].is_sent_start = True
    doc[1].is_sent_start = False
    doc = wp(doc)
    cleaned_words = [wp.model.clean_wp_token(t) for t in doc._.pytt_word_pieces_]
    cleaned_words = [w for w in cleaned_words if not is_special_token(w)]
    assert "".join(cleaned_words) == "".join(words)


def test_xlnet_weird_align(name, wp):
    if "xlnet" not in name.lower():
        return True
    text = "Well, i rented this movie and found out it realllllllly sucks."
    spacy_tokens = [
        "Well",
        ",",
        "i",
        "rented",
        "this",
        "movie",
        "and",
        "found",
        "out",
        "it",
        "realllllllly",
        "sucks",
        ".",
    ]
    spaces = [True] * len(spacy_tokens)
    spaces[0] = False
    spaces[-2] = False
    spaces[-1] = False
    doc = Doc(wp.vocab, words=spacy_tokens, spaces=spaces)
    doc[1].is_sent_start = False
    assert doc.text == text
    doc = wp(doc)
    assert doc._.pytt_word_pieces_[-2] == "</s>"
    assert doc._.pytt_word_pieces_[-1] == "<cls>"


def test_tokenizers_to_from_bytes(name):
    text = "hello world"
    tokenizer_cls = get_pytt_tokenizer(name)
    tokenizer = tokenizer_cls.from_pretrained(name)
    doc = tokenizer.tokenize(text)
    assert isinstance(doc, list) and len(doc)
    bytes_data = tokenizer.to_bytes(name)
    new_tokenizer = tokenizer_cls.blank().from_bytes(bytes_data)
    new_doc = new_tokenizer.tokenize(text)
    assert isinstance(new_doc, list) and len(new_doc)
    assert doc == new_doc



import pytest
import numpy
from numpy.testing import assert_equal
from spacy_pytorch_transformers.pipeline.wordpiecer import align_word_pieces
from spacy_pytorch_transformers.util import pad_batch, batch_by_length, lengths2mask


@pytest.mark.parametrize(
    "spacy_tokens,wp_tokens,expected_alignment",
    [
        (["a"], ["a"], [[0]]),
        (["a", "b", "c"], ["a", "b", "c"], [[0], [1], [2]]),
        (["ab", "c"], ["a", "b", "c"], [[0, 1], [2]]),
        (["a", "b", "c"], ["ab", "c"], [[0], [0], [1]]),
        (["ab", "cd"], ["a", "bc", "d"], [[0, 1], [1, 2]]),
        (["abcd"], ["ab", "cd"], [[0, 1]]),
        ([], [], []),
        (
            ["it", "realllllllly", "sucks", "."],
            ["it", "real", "ll", "ll", "ll", "ly", "suck", "s", "."],
            [[0], [1, 2, 3, 4, 5], [6, 7], [8]],
        ),
        (["Well", ",", "i"], ["Well", ",", "", "i"], [[0], [1, 2], [2, 3]]),
    ],
)
def test_align_word_pieces(spacy_tokens, wp_tokens, expected_alignment):
    output = align_word_pieces(spacy_tokens, wp_tokens)
    assert output == expected_alignment


@pytest.mark.parametrize(
    "lengths,max_words,expected",
    [
        ([1, 2, 2, 4], 4, [[3], [2], [0, 1]]),
        ([1, 2, 2, 4], 3, [[3], [2], [1], [0]]),
        ([4, 2, 2, 1], 4, [[0], [2], [1, 3]]),
        ([4, 4, 2, 2, 1], 6, [[1], [0], [2, 3, 4]]),
        ([10, 7, 2, 2, 1], 10, [[0], [1], [2, 3, 4]]),
    ],
)
def test_batch_by_length(lengths, max_words, expected):
    seqs = ["a" * length for length in lengths]
    batches = batch_by_length(seqs, max_words)
    assert batches == expected


@pytest.mark.parametrize(
    "lengths,expected_shape",
    [
        ([1, 2], (2, 2)),
        ([1, 2, 3], (3, 3)),
        ([0, 1], (2, 1)),
        ([1], (1, 1)),
        ([1, 5, 2, 4], (4, 5)),
    ],
)
def test_pad_batch(lengths, expected_shape):
    seqs = [numpy.ones((length,), dtype="f") for length in lengths]
    padded = pad_batch(seqs)
    for i, seq in enumerate(seqs):
        padded[i][padded[i] < 0] = 0
        assert padded[i].sum() == seq.sum()
        assert_equal(padded[i, : len(seq)], seq)
    assert padded.shape == expected_shape


@pytest.mark.parametrize(
    "lengths,width,expected_shape",
    [
        ([1, 2], 5, (2, 2, 5)),
        ([1, 2, 3], 2, (3, 3, 2)),
        ([0, 1], 3, (2, 1, 3)),
        ([1], 4, (1, 1, 4)),
        ([1, 5, 2, 4], 6, (4, 5, 6)),
    ],
)
def test_pad_batch_2d(lengths, width, expected_shape):
    seqs = [numpy.ones((length, width), dtype="f") for length in lengths]
    padded = pad_batch(seqs)
    for i, seq in enumerate(seqs):
        padded[i][padded[i] < 0] = 0
        assert padded[i].sum() == seq.sum()
        assert_equal(padded[i, : len(seq)], seq)
    assert padded.shape == expected_shape


@pytest.mark.parametrize("lengths", [[1], [1, 2], [5, 3, 1]])
def test_lengths2mask(lengths):
    mask = lengths2mask(lengths)
    assert mask.ndim == 1
    assert mask.sum() == sum(lengths)
    assert mask.size == max(lengths) * len(lengths)



import pytest
import numpy
from spacy_pytorch_transformers.activations import Activations, RaggedArray


def test_act_blank():
    acts = Activations.blank()
    assert acts.lh.data.size == 0
    assert acts.lh.lengths == []
    assert acts.po.data.size == 0
    assert acts.po.lengths == []
    assert not acts.has_lh
    assert not acts.has_po


@pytest.mark.parametrize(
    "extra_dims,lengths,pad_to,expected_shape",
    [
        ((), [1], 1, (1, 1)),
        ((), [1, 2], -1, (2, 2)),
        ((3,), [1, 2], -1, (2, 2, 3)),
        ((3,), [1, 2, 5], -1, (3, 5, 3)),
        ((3,), [1, 2, 5], 4, (3, 5, 3)),
    ],
)
def test_ragged_to_padded(extra_dims, lengths, pad_to, expected_shape):
    arr = RaggedArray(numpy.ones((sum(lengths),) + extra_dims), lengths)
    if pad_to > 1 and pad_to < max(lengths):
        with pytest.raises(ValueError):
            padded = arr.to_padded(to=pad_to)
    else:
        padded = arr.to_padded(to=pad_to)
        assert padded.shape == expected_shape


@pytest.mark.parametrize(
    "shape,lengths,expected_shape",
    [
        ((1, 2), [3], (3,)),
        ((2, 3), [3, 2], (5,)),
        ((2, 4, 4), [4, 2], (6, 4)),
        ((4, 5, 2), [5, 2, 1, 3], (11, 2)),
    ],
)
def test_ragged_from_padded(shape, lengths, expected_shape):
    padded = numpy.ones(shape, dtype="i")
    for i, length in enumerate(lengths):
        padded[i, length:] = 0
    ragged = RaggedArray.from_padded(padded, lengths)
    assert ragged.data.shape == expected_shape
    assert ragged.data.sum() == padded.sum()


@pytest.mark.parametrize(
    "shape,lengths,expected_shape",
    [
        ((1, 2), [3], (3,)),
        ((2, 2), [3, 2], (5,)),
        ((2, 2, 4), [3, 2], (5, 4)),
        ((4, 2, 2), [5, 2, 1, 3], (11, 2)),
    ],
)
def test_ragged_from_truncated(shape, lengths, expected_shape):
    truncated = numpy.ones(shape, dtype="i")
    for i, length in enumerate(lengths):
        truncated[i, length:] = 0
    ragged = RaggedArray.from_truncated(truncated, lengths)
    assert ragged.data.shape == expected_shape
    assert ragged.data.sum() == truncated.sum()



import numpy
import pytest
from thinc.neural.optimizers import Adam

from spacy_pytorch_transformers.activations import Activations, RaggedArray
from spacy_pytorch_transformers.util import get_pytt_tokenizer


@pytest.fixture
def tokenizer(name):
    return get_pytt_tokenizer(name).from_pretrained(name)


@pytest.fixture
def ids(tokenizer):
    text = "the cat sat on the mat"
    return numpy.array(tokenizer.encode(text), dtype=numpy.int_)


@pytest.fixture
def inputs(ids):
    return RaggedArray(ids, [len(ids)])


@pytest.fixture
def model(nlp):
    return nlp.get_pipe("pytt_tok2vec").model._model


def test_wrapper_from_pretrained(name, model, inputs):
    outputs, backprop = model.begin_update(inputs)
    assert outputs.has_lh
    optimizer = Adam(model.ops, 0.001)
    d_outputs = Activations(outputs.lh, RaggedArray.blank())
    backprop(d_outputs, sgd=optimizer)



from typing import Tuple, Callable, List, Optional
from thinc.api import wrap, layerize, chain, flatten_add_lengths, with_getitem
from thinc.t2v import Pooling, mean_pool
from thinc.v2v import Softmax, Affine, Model
from thinc.neural.util import get_array_module
from spacy.tokens import Span, Doc
import numpy

from .wrapper import PyTT_Wrapper
from .util import Array, Dropout, Optimizer
from .util import batch_by_length, flatten_list, is_class_token
from .util import get_segment_ids, is_special_token
from .activations import Activations as Acts
from .activations import RaggedArray


REGISTRY = {}


def register_model(name: str, model=None):
    """Decorator to register a model."""
    global REGISTRY
    if model is not None:
        REGISTRY[name] = model
        return model

    def do_registration(model):
        REGISTRY[name] = model
        return model

    return do_registration


def get_model_function(name: str):
    """Get a model creation function from the registry by name."""
    if name not in REGISTRY:
        names = ", ".join(sorted(REGISTRY.keys()))
        raise KeyError(f"Model {name} not found in registry. Available names: {names}")
    return REGISTRY[name]


@register_model("tok2vec_per_sentence")
def tok2vec_per_sentence(pytt_model, cfg):
    max_words = cfg.get("words_per_batch", 1000)
    name = cfg["pytt_name"]

    model = foreach_sentence(
        chain(get_word_pieces(name), with_length_batching(pytt_model, max_words))
    )
    return model


@register_model("softmax_tanh_class_vector")
def softmax_tanh_class_vector(nr_class, *, exclusive_classes=True, **cfg):
    """Select features from the class-vectors from the last hidden state,
    mean-pool them, apply a tanh-activated hidden layer, and then softmax-activated
    output layer to produce one vector per document.
    The gradients of the class vectors are incremented in the backward pass,
    to allow fine-tuning.
    """
    width = cfg["token_vector_width"]
    return chain(
        get_pytt_class_tokens,
        flatten_add_lengths,
        with_getitem(0, chain(Affine(width, width), tanh)),
        Pooling(mean_pool),
        Softmax(2, width),
    )


@register_model("softmax_class_vector")
def softmax_class_vector(nr_class, *, exclusive_classes=True, **cfg):
    """Select features from the class-vectors from the last hidden state,
    mean-pool them, and apply a softmax-activated hidden layer to produce one
    vector per document. The gradients of the class vectors are incremented
    in the backward pass, to allow fine-tuning.
    """
    width = cfg["token_vector_width"]
    return chain(
        get_pytt_class_tokens,
        flatten_add_lengths,
        Pooling(mean_pool),
        Softmax(2, width),
    )


@register_model("softmax_last_hidden")
def softmax_last_hidden(nr_class, *, exclusive_classes=True, **cfg):
    """Select features from the class-vectors from the last hidden state,
    mean-pool them, and softmax to produce one vector per document.
    The gradients of the class vectors are incremented in the backward pass,
    to allow fine-tuning.
    """
    width = cfg["token_vector_width"]
    return chain(
        get_pytt_last_hidden, flatten_add_lengths, Pooling(mean_pool), Softmax(2, width)
    )


@register_model("softmax_pooler_output")
def softmax_pooler_output(nr_class, *, exclusive_classes=True, **cfg):
    """Select features from the pooler output, (if necessary) mean-pool them
    to produce one vector per item, and then softmax them.
    The gradients of the class vectors are incremented in the backward pass,
    to allow fine-tuning.
    """
    return chain(
        get_pytt_pooler_output,
        flatten_add_lengths,
        with_getitem(0, Softmax(nr_class, cfg["token_vector_width"])),
        Pooling(mean_pool),
    )


@layerize
def get_pytt_class_tokens(docs, drop=0.0):
    """Output a List[array], where the array is the class vector
    for each sentence in the document. To backprop, we increment the values
    in the doc._.pytt_d_last_hidden_state array.
    """
    xp = get_array_module(docs[0]._.pytt_last_hidden_state)
    outputs = []
    doc_class_tokens = []
    for doc in docs:
        class_tokens = []
        for i, wp in enumerate(doc._.pytt_word_pieces_):
            if is_class_token(wp):
                class_tokens.append(i)
        doc_class_tokens.append(xp.array(class_tokens, dtype="i"))
        wp_tensor = doc._.pytt_last_hidden_state
        outputs.append(wp_tensor[doc_class_tokens[-1]])

    def backprop_pytt_class_tokens(d_outputs, sgd=None):
        for doc, class_tokens, dY in zip(docs, doc_class_tokens, d_outputs):
            if doc._.pytt_d_last_hidden_state.size == 0:
                xp = get_array_module(doc._.pytt_last_hidden_state)
                grads = xp.zeros(doc._.pytt_last_hidden_state.shape, dtype="f")
                doc._.pytt_d_last_hidden_state = grads
            doc._.pytt_d_last_hidden_state[class_tokens] += dY
        return None

    return outputs, backprop_pytt_class_tokens


get_pytt_class_tokens.name = "get_pytt_class_tokens"


@layerize
def get_pytt_pooler_output(docs, drop=0.0):
    """Output a List[array], where the array is the class vector
    for each sentence in the document. To backprop, we increment the values
    in the doc._.pytt_d_last_hidden_state array.
    """
    for doc in docs:
        if doc._.pytt_pooler_output is None:
            raise ValueError(
                "Pooler output unset. Perhaps you're using the wrong architecture? "
                "The BERT model provides a pooler output, but XLNet doesn't. "
                "You might need to set 'architecture': 'softmax_class_vector' "
                "instead."
            )
    outputs = [doc._.pytt_pooler_output for doc in docs]

    def backprop_pytt_pooler_output(d_outputs, sgd=None):
        for doc, dY in zip(docs, d_outputs):
            if doc._.pytt_d_pooler_output.size == 0:
                xp = get_array_module(doc._.pytt_pooler_output)
                grads = xp.zeros(doc._.pytt_pooler_output.shape, dtype="f")
                doc._.pytt_d_pooler_output = grads
            doc._.pytt_d_pooler_output += dY
        return None

    return outputs, backprop_pytt_pooler_output


get_pytt_pooler_output.name = "get_pytt_pooler_output"


@layerize
def get_pytt_last_hidden(docs, drop=0.0):
    """Output a List[array], where the array is the last hidden vector vector
    for each document. To backprop, we increment the values
    in the doc._.pytt_d_last_hidden_state array.
    """
    outputs = [doc._.pytt_last_hidden_state for doc in docs]
    for out in outputs:
        assert out is not None
        assert out.size != 0

    def backprop_pytt_last_hidden(d_outputs, sgd=None):
        for doc, d_lh in zip(docs, d_outputs):
            xp = get_array_module(d_lh)
            shape = d_lh.shape
            dtype = d_lh.dtype
            if doc._.pytt_d_last_hidden_state.size == 0:
                doc._.pytt_d_last_hidden_state = xp.zeros(shape, dtype=dtype)
            doc._.pytt_d_last_hidden_state += d_lh
        return None

    return outputs, backprop_pytt_last_hidden


get_pytt_last_hidden.name = "get_pytt_last_hidden"


@layerize
def softmax(X, drop=0.0):
    ops = Model.ops
    Y = ops.softmax(X)

    def backprop_softmax(dY, sgd=None):
        dX = ops.backprop_softmax(Y, dY)
        return dX

    return Y, backprop_softmax


@layerize
def tanh(X, drop=0.0):
    xp = get_array_module(X)
    Y = xp.tanh(X)

    def backprop_tanh(dY, sgd=None):
        one = Y.dtype.type(1)
        dX = dY * (one - Y * Y)
        return dX

    return Y, backprop_tanh


def get_word_pieces(pytt_name):
    def get_features_forward(sents, drop=0.0):
        assert isinstance(sents[0], Span)
        ids = []
        segment_ids = []
        lengths = []
        for sent in sents:
            wordpieces = sent._.pytt_word_pieces
            # This is a bit convoluted, but we need the lengths without any
            # separator tokens or class tokens. pytt_segments gives Span objects.
            seg_lengths = [
                len([w for w in seg._.pytt_word_pieces_ if not is_special_token(w)])
                for seg in sent._.pytt_segments
            ]
            if wordpieces:
                ids.extend(wordpieces)
                lengths.append(len(wordpieces))
                sent_seg_ids = get_segment_ids(pytt_name, *seg_lengths)
                segment_ids.extend(sent_seg_ids)
                assert len(wordpieces) == len(sent_seg_ids), (
                    sent._.pytt_word_pieces_,
                    seg_lengths,
                    len(wordpieces),
                    len(sent_seg_ids),
                )
            else:
                lengths.append(0)
        assert len(ids) == len(segment_ids), (len(ids), len(segment_ids))
        features = numpy.array(list(zip(ids, segment_ids)), dtype=numpy.int_)
        assert features.shape[0] == sum(lengths), (features.shape, sum(lengths))
        return RaggedArray(features, lengths), None

    return layerize(get_features_forward, name="get_features_forward")


def with_length_batching(model: PyTT_Wrapper, max_words: int) -> PyTT_Wrapper:
    ops = model.ops

    def apply_model_to_batches(
        inputs: RaggedArray, drop: Dropout = 0.0
    ) -> Tuple[Acts, Callable]:
        if max_words == 0 or inputs.data.shape[0] < max_words:
            return model.begin_update(inputs, drop=drop)
        Xs: List[Array] = ops.unflatten(inputs.data, inputs.lengths)
        outputs = None
        backprops = []
        index2rows = {}
        start = 0
        # Map each index to the slice of rows in the flattened data it refers to.
        for i, length in enumerate(inputs.lengths):
            index2rows[i] = [start + j for j in range(length)]
            start += length
        total_rows = sum(inputs.lengths)
        for indices in batch_by_length(Xs, max_words):
            X: Array = inputs.xp.concatenate([Xs[i] for i in indices])
            lengths = [inputs.lengths[i] for i in indices]
            Y, get_dX = model.begin_update(RaggedArray(X, lengths), drop=drop)
            if outputs is None:
                lh_shape = (total_rows, Y.lh.data.shape[-1])
                po_shape = (len(inputs.lengths), Y.po.data.shape[-1])
                outputs = Acts(
                    RaggedArray(Y.lh.xp.zeros(lh_shape, dtype="f"), inputs.lengths),
                    RaggedArray(
                        Y.po.xp.zeros(po_shape, dtype="f"), [1 for _ in inputs.lengths]
                    ),
                )
            lh_rows = []
            po_rows = []
            for index in indices:
                lh_rows.extend(index2rows[index])
                po_rows.append(index)
            lh_rows = outputs.xp.array(lh_rows, dtype="i")
            po_rows = outputs.xp.array(po_rows, dtype="i")
            outputs.lh.data[lh_rows] = Y.lh.data
            if outputs.has_po and po_rows.size:
                outputs.po.data[po_rows] = Y.po.data
            backprops.append((get_dX, lh_rows, po_rows, lengths))

        def backprop_batched(d_outputs: Acts, sgd: Optimizer = None):
            for get_dX, lh_rows, po_rows, lengths in backprops:
                if d_outputs.has_lh:
                    d_lh = d_outputs.lh.data[lh_rows]
                    lh_lengths = lengths
                else:
                    d_lh = d_outputs.lh.data
                    lh_lengths = []
                if d_outputs.has_po:
                    d_po = d_outputs.po.data[po_rows]
                    po_lengths = [1 for _ in lengths]
                else:
                    d_po = d_outputs.po.data
                    po_lengths = []
                dY = Acts(RaggedArray(d_lh, lh_lengths), RaggedArray(d_po, po_lengths))
                dX = get_dX(dY, sgd=sgd)
                assert dX is None
            return None

        return outputs, backprop_batched

    return wrap(apply_model_to_batches, model)


def foreach_sentence(layer: Model, drop_factor: float = 1.0) -> Model:
    """Map a layer across sentences (assumes spaCy-esque .sents interface)"""

    def sentence_fwd(docs: List[Doc], drop: Dropout = 0.0) -> Tuple[Acts, Callable]:
        if not all(doc.is_sentenced for doc in docs):
            return layer.begin_update([d[:] for d in docs], drop=drop)
        sents = flatten_list([list(doc.sents) for doc in docs])
        words_per_doc = [len(d._.pytt_word_pieces) for d in docs]
        words_per_sent = [len(s._.pytt_word_pieces) for s in sents]
        sents_per_doc = [len(list(d.sents)) for d in docs]
        assert sum(words_per_doc) == sum(words_per_sent)
        acts, bp_acts = layer.begin_update(sents, drop=drop)
        # To go from "per sentence" activations to "per doc" activations, we
        # just have to tell it where the sequences end.
        acts.lh.lengths = words_per_doc
        acts.po.lengths = sents_per_doc

        def sentence_bwd(d_acts: Acts, sgd: Optional[Optimizer] = None) -> None:
            assert isinstance(d_acts, Acts)
            # Translate back to the per-sentence activations
            if d_acts.has_lh:
                assert d_acts.lh.data.shape[0] == sum(d_acts.lh.lengths)
                assert d_acts.lh.lengths == words_per_doc
            d_acts.lh.lengths = words_per_sent
            d_acts.po.lengths = [1 for _ in words_per_sent]
            d_ids = bp_acts(d_acts, sgd=sgd)
            if not (d_ids is None or all(ds is None for ds in d_ids)):
                raise ValueError("Expected gradient of sentence to be None")
            return d_ids

        return acts, sentence_bwd

    return wrap(sentence_fwd, layer)



from typing import Any, List
from thinc.neural.ops import get_array_module
from spacy.pipeline import Pipe
from spacy.tokens import Doc
from spacy.vocab import Vocab
from spacy.util import minibatch

from ..wrapper import PyTT_Wrapper
from ..model_registry import get_model_function
from ..activations import Activations, RaggedArray
from ..util import get_pytt_config, get_pytt_model, get_sents


class PyTT_TokenVectorEncoder(Pipe):
    """spaCy pipeline component to use PyTorch-Transformers models.

    The component assigns the output of the transformer to the `doc._.pytt_outputs`
    extension attribute. We also calculate an alignment between the word-piece
    tokens and the spaCy tokenization, so that we can use the last hidden states
    to set the doc.tensor attribute. When multiple word-piece tokens align to
    the same spaCy token, the spaCy token receives the sum of their values.
    """

    name = "pytt_tok2vec"

    @classmethod
    def from_nlp(cls, nlp, **cfg):
        """Factory to add to Language.factories via entry point."""
        return cls(nlp.vocab, **cfg)

    @classmethod
    def from_pretrained(cls, vocab: Vocab, name: str, **cfg):
        """Create a PyTT_TokenVectorEncoder instance using pre-trained weights
        from a PyTorch Transformer model, even if it's not installed as a
        spaCy package.

        vocab (spacy.vocab.Vocab): The spaCy vocab to use.
        name (unicode): Name of pre-trained model, e.g. 'bert-base-uncased'.
        RETURNS (PyTT_TokenVectorEncoder): The token vector encoder.
        """
        cfg["pytt_name"] = name
        model = cls.Model(from_pretrained=True, **cfg)
        cfg["pytt_config"] = dict(model._model.pytt_model.config.to_dict())
        self = cls(vocab, model=model, **cfg)
        return self

    @classmethod
    def Model(cls, **cfg) -> Any:
        """Create an instance of `PyTT_Wrapper`, which holds the
        PyTorch-Transformers model.

        **cfg: Optional config parameters.
        RETURNS (thinc.neural.Model): The wrapped model.
        """
        name = cfg.get("pytt_name")
        if not name:
            raise ValueError("Need pytt_name argument, e.g. 'bert-base-uncased'")
        if cfg.get("from_pretrained"):
            pytt_wrap = PyTT_Wrapper.from_pretrained(name)
        else:
            pytt_config = cfg["pytt_config"]
            # Work around floating point limitation in ujson:
            # If we have the setting cfg["pytt_config"]["layer_norm_eps"] as 0,
            # that's because of misprecision in serializing. Fix that.
            pytt_config["layer_norm_eps"] = 1e-12
            config_cls = get_pytt_config(name)
            model_cls = get_pytt_model(name)
            # Need to match the name their constructor expects.
            if "vocab_size" in cfg["pytt_config"]:
                vocab_size = cfg["pytt_config"]["vocab_size"]
                cfg["pytt_config"]["vocab_size_or_config_json_file"] = vocab_size
            pytt_wrap = PyTT_Wrapper(
                name, pytt_config, model_cls(config_cls(**pytt_config))
            )
        make_model = get_model_function(cfg.get("architecture", "tok2vec_per_sentence"))
        model = make_model(pytt_wrap, cfg)
        setattr(model, "nO", pytt_wrap.nO)
        setattr(model, "_model", pytt_wrap)
        return model

    def __init__(self, vocab, model=True, **cfg):
        """Initialize the component.

        model (thinc.neural.Model / True): The component's model or `True` if
            not initialized yet.
        **cfg: Optional config parameters.
        """
        self.vocab = vocab
        self.model = model
        self.cfg = cfg

    @property
    def token_vector_width(self):
        return self.model._model.nO

    @property
    def pytt_model(self):
        return self.model._model.pytt_model

    def __call__(self, doc):
        """Process a Doc and assign the extracted features.

        doc (spacy.tokens.Doc): The Doc to process.
        RETURNS (spacy.tokens.Doc): The processed Doc.
        """
        self.require_model()
        outputs = self.predict([doc])
        self.set_annotations([doc], outputs)
        return doc

    def pipe(self, stream, batch_size=128):
        """Process Doc objects as a stream and assign the extracted features.

        stream (iterable): A stream of Doc objects.
        batch_size (int): The number of texts to buffer.
        YIELDS (spacy.tokens.Doc): Processed Docs in order.
        """
        for docs in minibatch(stream, size=batch_size):
            docs = list(docs)
            outputs = self.predict(docs)
            self.set_annotations(docs, outputs)
            for doc in docs:
                yield doc

    def begin_update(self, docs, drop=None, **cfg):
        """Get the predictions and a callback to complete the gradient update.
        This is only used internally within PyTT_Language.update.
        """
        outputs, backprop = self.model.begin_update(docs, drop=drop)

        def finish_update(docs, sgd=None):
            assert len(docs)
            d_lh = []
            d_po = []
            lh_lengths = []
            po_lengths = []
            for doc in docs:
                d_lh.append(doc._.pytt_d_last_hidden_state)
                d_po.append(doc._.pytt_d_pooler_output)
                lh_lengths.append(doc._.pytt_d_last_hidden_state.shape[0])
                po_lengths.append(doc._.pytt_d_pooler_output.shape[0])
            xp = self.model.ops.xp
            gradients = Activations(
                RaggedArray(xp.vstack(d_lh), lh_lengths),
                RaggedArray(xp.vstack(d_po), po_lengths),
            )
            backprop(gradients, sgd=sgd)
            for doc in docs:
                doc._.pytt_d_last_hidden_state.fill(0)
                doc._.pytt_d_pooler_output.fill(0)
            return None

        return outputs, finish_update

    def predict(self, docs):
        """Run the transformer model on a batch of docs and return the
        extracted features.

        docs (iterable): A batch of Docs to process.
        RETURNS (list): A list of Activations objects, one per doc.
        """
        return self.model.predict(docs)

    def set_annotations(self, docs: List[Doc], activations: Activations):
        """Assign the extracted features to the Doc objects and overwrite the
        vector and similarity hooks.

        docs (iterable): A batch of `Doc` objects.
        activations (iterable): A batch of activations.
        """
        xp = activations.xp
        for i, doc in enumerate(docs):
            # Make it 2d -- acts are always 3d, to represent batch size.
            wp_tensor = activations.lh.get(i)
            doc.tensor = self.model.ops.allocate((len(doc), self.model.nO))
            doc._.pytt_last_hidden_state = wp_tensor
            if activations.has_po:
                pooler_output = activations.po.get(i)
                doc._.pytt_pooler_output = pooler_output
            doc._.pytt_d_last_hidden_state = xp.zeros((0, 0), dtype=wp_tensor.dtype)
            doc._.pytt_d_pooler_output = xp.zeros((0, 0), dtype=wp_tensor.dtype)
            doc._.pytt_d_all_hidden_states = []
            doc._.pytt_d_all_attentions = []
            if wp_tensor.shape != (len(doc._.pytt_word_pieces), self.model.nO):
                raise ValueError(
                    "Mismatch between tensor shape and word pieces. This usually "
                    "means we did something wrong in the sentence reshaping, "
                    "or possibly finding the separator tokens."
                )
            # Count how often each word-piece token is represented. This allows
            # a weighted sum, so that we can make sure doc.tensor.sum()
            # equals wp_tensor.sum(). Do this with sensitivity to boundary tokens
            wp_rows, align_sizes = _get_boundary_sensitive_alignment(doc)
            wp_weighted = wp_tensor / xp.array(align_sizes, dtype="f").reshape((-1, 1))
            # TODO: Obviously incrementing the rows individually is bad. How
            # to do in one shot without blowing up the memory?
            for i, word_piece_slice in enumerate(wp_rows):
                for j in word_piece_slice:
                    doc.tensor[i] += wp_weighted[j]
            doc.user_hooks["vector"] = get_doc_vector_via_tensor
            doc.user_span_hooks["vector"] = get_span_vector_via_tensor
            doc.user_token_hooks["vector"] = get_token_vector_via_tensor
            doc.user_hooks["similarity"] = get_similarity_via_tensor
            doc.user_span_hooks["similarity"] = get_similarity_via_tensor
            doc.user_token_hooks["similarity"] = get_similarity_via_tensor


def _get_boundary_sensitive_alignment(doc):
    align_sizes = [0 for _ in range(len(doc._.pytt_word_pieces))]
    wp_rows = []
    for word_piece_slice in doc._.pytt_alignment:
        wp_rows.append(list(word_piece_slice))
        for i in word_piece_slice:
            align_sizes[i] += 1
    # To make this weighting work, we "align" the boundary tokens against
    # every token in their sentence. The boundary tokens are otherwise
    # unaligned, which is how we identify them.
    for sent in get_sents(doc):
        offset = sent._.pytt_start
        for i in range(len(sent._.pytt_word_pieces)):
            if align_sizes[offset + i] == 0:
                align_sizes[offset + i] = len(sent)
                for tok in sent:
                    wp_rows[tok.i].append(offset + i)
    return wp_rows, align_sizes


def get_doc_vector_via_tensor(doc):
    return doc.tensor.sum(axis=0)


def get_span_vector_via_tensor(span):
    return span.doc.tensor[span.start : span.end].sum(axis=0)


def get_token_vector_via_tensor(token):
    return token.doc.tensor[token.i]


def get_similarity_via_tensor(doc1, doc2):
    v1 = doc1.vector
    v2 = doc2.vector
    xp = get_array_module(v1)
    return xp.dot(v1, v2) / (doc1.vector_norm * doc2.vector_norm)



from contextlib import contextmanager
from pathlib import Path
import tempfile
import shutil
import numpy


@contextmanager
def make_tempdir():
    d = Path(tempfile.mkdtemp())
    yield d
    shutil.rmtree(str(d))


def is_valid_tensor(tensor):
    return tensor is not None and numpy.nonzero(tensor) and tensor.size != 0



import os
import io
from setuptools import setup, find_packages


def setup_package():
    package_name = "spacy_pytorch_transformers"
    root = os.path.abspath(os.path.dirname(__file__))

    # Read in package meta from about.py
    about_path = os.path.join(root, package_name, "about.py")
    with io.open(about_path, encoding="utf8") as f:
        about = {}
        exec(f.read(), about)

    # Get readme
    readme_path = os.path.join(root, "README.md")
    with io.open(readme_path, encoding="utf8") as f:
        readme = f.read()

    setup(
        name="spacy-pytorch-transformers",
        description=about["__summary__"],
        long_description=readme,
        long_description_content_type="text/markdown",
        author=about["__author__"],
        author_email=about["__email__"],
        url=about["__uri__"],
        version=about["__version__"],
        license=about["__license__"],
        packages=find_packages(),
        install_requires=[
            "spacy>=2.1.7,<2.2.0",
            "pytorch_transformers>=1.0.0,<1.1.0",
            "torch>=1.0.0",
            "torchcontrib>=0.0.2,<0.1.0",
            "srsly>=0.0.7,<1.1.0",
            "ftfy>=5.0.0,<6.0.0",
            "dataclasses>=0.6,<0.7; python_version < '3.7'",
        ],
        python_requires=">=3.6",
        extras_require={
            "cuda": ["thinc_gpu_ops>=0.0.1,<0.1.0", "cupy>=5.0.0b4"],
            "cuda80": ["thinc_gpu_ops>=0.0.1,<0.1.0", "cupy-cuda80>=5.0.0b4"],
            "cuda90": ["thinc_gpu_ops>=0.0.1,<0.1.0", "cupy-cuda90>=5.0.0b4"],
            "cuda91": ["thinc_gpu_ops>=0.0.1,<0.1.0", "cupy-cuda91>=5.0.0b4"],
            "cuda92": ["thinc_gpu_ops>=0.0.1,<0.1.0", "cupy-cuda92>=5.0.0b4"],
            "cuda100": ["thinc_gpu_ops>=0.0.1,<0.1.0", "cupy-cuda100>=5.0.0b4"],
        },
        entry_points={
            "spacy_factories": [
                "pytt_wordpiecer = spacy_pytorch_transformers:PyTT_WordPiecer.from_nlp",
                "pytt_tok2vec = spacy_pytorch_transformers:PyTT_TokenVectorEncoder.from_nlp",
                "pytt_textcat = spacy_pytorch_transformers:PyTT_TextCategorizer.from_nlp",
            ],
            "spacy_languages": ["pytt = spacy_pytorch_transformers:PyTT_Language"],
        },
        classifiers=[
            "Development Status :: 4 - Beta",
            "Intended Audience :: Developers",
            "Topic :: Scientific/Engineering :: Artificial Intelligence",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
        ],
        zip_safe=False,
    )


if __name__ == "__main__":
    setup_package()



"""Adjust the initialization and serialization of the PyTorch Transformers
tokenizers, so that they work more nicely with spaCy. Specifically, the PyTT
classes take file paths as arguments to their __init__, which means we can't
easily use to_bytes() and from_bytes() with them.

Additionally, provide a .clean_text() method that is used to match the preprocessing,
so that we can get the alignment working.
"""
from collections import OrderedDict
import spacy
import ftfy
import srsly
import regex
import sentencepiece
from pathlib import Path
import unicodedata
import re

import pytorch_transformers as pytt
from pytorch_transformers.tokenization_gpt2 import bytes_to_unicode
from pytorch_transformers.tokenization_bert import BasicTokenizer, WordpieceTokenizer


BASE_CLASS_FIELDS = [
    "_bos_token",
    "_eos_token",
    "_unk_token",
    "_sep_token",
    "_pad_token",
    "_cls_token",
    "_mask_token",
    "_additional_special_tokens",
    "max_len",
    "added_tokens_encoder",
    "added_tokens_decoder",
]


class SerializationMixin:
    """Provide generic serialization methods, for compatibility with spaCy.
    These expect the tokenizer subclass to provide the following:

    * serialization_fields (List[str]): List of attributes to serialize. All
        attributes should have json-serializable values.
    * finish_deserializing(): A function to be called after from_bytes(),
        to finish setting up the instance.
    """

    def prepare_for_serialization(self):
        pass

    def from_bytes(self, bytes_data, exclude=tuple(), **kwargs):
        msg = srsly.msgpack_loads(bytes_data)
        for field in self.serialization_fields:
            setattr(self, field, msg[field])
        self.finish_deserializing()
        return self

    def to_bytes(self, exclude=tuple(), **kwargs):
        self.prepare_for_serialization()
        msg = OrderedDict()
        for field in self.serialization_fields:
            msg[field] = getattr(self, field, None)
        return srsly.msgpack_dumps(msg)

    def from_disk(self, path, exclude=tuple(), **kwargs):
        with (path / "pytt_tokenizer.msg").open("rb") as file_:
            data = file_.read()
        return self.from_bytes(data, **kwargs)

    def to_disk(self, path, exclude=tuple(), **kwargs):
        self.prepare_for_serialization()
        data = self.to_bytes(**kwargs)
        with (path / "pytt_tokenizer.msg").open("wb") as file_:
            file_.write(data)


class SerializableBertTokenizer(pytt.BertTokenizer, SerializationMixin):
    serialization_fields = list(BASE_CLASS_FIELDS) + [
        "vocab",
        "do_basic_tokenize",
        "do_lower_case",
        "never_split",
        "tokenize_chinese_chars",
    ]

    @classmethod
    def blank(cls):
        self = cls.__new__(cls)
        for field in self.serialization_fields:
            setattr(self, field, None)
        self.ids_to_tokens = None
        self.basic_tokenizer = None
        self.wordpiece_tokenizer = None
        return self

    def prepare_for_serialization(self):
        if self.basic_tokenizer is not None:
            self.do_lower_case = self.basic_tokenizer.do_lower_case
            self.never_split = self.basic_tokenizer.never_split
            self.tokenize_chinese_chars = self.basic_tokenizer.tokenize_chinese_chars

    def finish_deserializing(self):
        self.ids_to_tokens = OrderedDict(
            [(ids, tok) for tok, ids in self.vocab.items()]
        )
        if self.do_basic_tokenize:
            self.basic_tokenizer = BasicTokenizer(
                do_lower_case=self.do_lower_case,
                never_split=self.never_split,
                tokenize_chinese_chars=self.tokenize_chinese_chars,
            )
        self.wordpiece_tokenizer = WordpieceTokenizer(
            vocab=self.vocab, unk_token=self.unk_token
        )

    def clean_token(self, text):
        if self.do_basic_tokenize:
            text = self.basic_tokenizer._clean_text(text)
        text = text.strip()
        return clean_accents(text)

    def clean_wp_token(self, token):
        return token.replace("##", "", 1).strip()

    def add_special_tokens(self, segments):
        output = []
        for segment in segments:
            output.extend(segment)
            if segment:
                output.append(self.sep_token)
        if output:
            # If we otherwise would have an empty output, don't add cls
            output.insert(0, self.cls_token)
        return output

    def fix_alignment(self, segments):
        """Turn a nested segment alignment into an alignment for the whole input,
        by offsetting and accounting for special tokens."""
        offset = 0
        output = []
        for segment in segments:
            if segment:
                offset += 1
            seen = set()
            for idx_group in segment:
                output.append([idx + offset for idx in idx_group])
                seen.update({idx for idx in idx_group})
            offset += len(seen)
        return output


class SerializableGPT2Tokenizer(pytt.GPT2Tokenizer, SerializationMixin):
    serialization_fields = list(BASE_CLASS_FIELDS) + [
        "encoder",
        "_bpe_ranks",
        "errors",
        "_regex_pattern",
    ]

    @classmethod
    def blank(cls):
        self = cls.__new__(cls)
        for field in self.serialization_fields:
            setattr(self, field, None)
        self.byte_encoder = None
        self.byte_decoder = None
        self.bpe_ranks = {}
        self.cache = None
        self.pat = None
        return self

    def finish_deserializing(self):
        self.bpe_ranks = deserialize_bpe_ranks(self._bpe_ranks)
        self.decoder = {v: k for k, v in self.encoder.items()}
        self.byte_encoder = bytes_to_unicode()
        self.byte_decoder = {v: k for k, v in self.byte_encoder.items()}
        self.cache = {}
        self.pat = regex.compile(self._regex_pattern, flags=regex.V0)

    def prepare_for_serialization(self):
        self._regex_pattern = self.pat.pattern
        self._bpe_ranks = serialize_bpe_ranks(self.bpe_ranks)

    def clean_token(self, text):
        text = clean_extended_unicode(text)
        return text.strip()

    def clean_wp_token(self, text):
        text = text.replace("\u0120", "", 1)
        text = text.replace("\u010a", "", 1)
        text = ftfy.fix_text(text)
        text = clean_extended_unicode(text)
        return text.strip()

    def add_special_tokens(self, segments):
        output = []
        for segment in segments:
            if segment:
                output.append(self.bos_token)
            output.extend(segment)
            if segment:
                output.append(self.eos_token)
        return output

    def fix_alignment(self, segments):
        """Turn a nested segment alignment into an alignment for the whole input,
        by offsetting and accounting for special tokens."""
        offset = 0
        output = []
        for segment in segments:
            if segment:
                offset += 1
            seen = set()
            for idx_group in segment:
                output.append([idx + offset for idx in idx_group])
                seen.update({idx for idx in idx_group})
            offset += len(seen)
            if segment:
                offset += 1
        return output


class SerializableXLMTokenizer(pytt.XLMTokenizer, SerializationMixin):
    _replace_re = re.compile(r"[\s\.\-`'\";]+")
    serialization_fields = list(BASE_CLASS_FIELDS) + ["encoder", "_bpe_ranks"]

    @classmethod
    def blank(cls):
        self = cls.__new__(cls)
        for field in self.serialization_fields:
            setattr(self, field, None)
        self.nlp = None
        self.fix_text = None
        self.cache = None
        self.decoder = {}
        self.bpe_ranks = {}
        return self

    def finish_deserializing(self):
        self.bpe_ranks = deserialize_bpe_ranks(self._bpe_ranks)
        self.nlp = spacy.blank("en")
        self.fix_text = ftfy.fix_text
        self.cache = {}
        self.decoder = {v: k for k, v in self.encoder.items()}

    def prepare_for_serialization(self):
        self._bpe_ranks = serialize_bpe_ranks(self.bpe_ranks)

    def clean_token(self, text):
        # Model seems to just strip out all unicode so we need to do this, too,
        # instead of calling clean_accents etc.
        text = ftfy.fix_text(text)
        text = re.sub(r"&(#\d+|#x[a-f\d]+)", "", text)  # malformed HTML entities
        text = clean_extended_unicode(text)
        text = self._replace_re.sub("", text)
        return text.strip()

    def clean_wp_token(self, text):
        text = ftfy.fix_text(text)
        text = clean_extended_unicode(text)
        text = self._replace_re.sub("", text)
        return text.replace("</w>", "").strip()

    def add_special_tokens(self, segments):
        # See https://github.com/facebookresearch/XLM/issues/113
        output = []
        for segment in segments:
            if segment:
                output.append(self.bos_token)
            output.extend(segment)
            if segment:
                output.append(self.eos_token)
        return output

    def fix_alignment(self, segments):
        """Turn a nested segment alignment into an alignment for the whole input,
        by offsetting and accounting for special tokens."""
        offset = 0
        output = []
        for segment in segments:
            if segment:
                offset += 1
            seen = set()
            for idx_group in segment:
                output.append([idx + offset for idx in idx_group])
                seen.update({idx for idx in idx_group})
            offset += len(seen)
            if segment:
                offset += 1
        return output


class SerializableXLNetTokenizer(pytt.XLNetTokenizer, SerializationMixin):
    _replace_re = re.compile(r"[\s'\";]+")
    _replacements = [("º", "o"), *zip("⁰¹²³⁴⁵⁶⁷⁸⁹", "0123456789")]
    serialization_fields = list(BASE_CLASS_FIELDS) + [
        "do_lower_case",
        "remove_space",
        "keep_accents",
        "vocab_bytes",
    ]

    @classmethod
    def blank(cls):
        self = cls.__new__(cls)
        for field in self.serialization_fields:
            setattr(self, field, None)
        self.sp_model = None
        return self

    def prepare_for_serialization(self):
        if hasattr(self, "vocab_file"):
            vocab_path = Path(self.vocab_file)
            with vocab_path.open("rb") as f:
                self.vocab_bytes = f.read()

    def finish_deserializing(self):
        self.sp_model = sentencepiece.SentencePieceProcessor()
        self.sp_model.LoadFromSerializedProto(self.vocab_bytes)

    def clean_token(self, text):
        text = clean_fractions(text)
        text = clean_accents(text)
        for a, b in self._replacements:
            text = text.replace(a, b)
        text = clean_extended_unicode(text)
        text = self._replace_re.sub("", text)
        return text.strip()

    def clean_wp_token(self, text):
        # Note: The special control character is \u2581
        text = clean_accents(text)
        for a, b in self._replacements:
            text = text.replace(a, b)
        text = clean_extended_unicode(text)
        text = self._replace_re.sub("", text)
        return text.strip()

    def add_special_tokens(self, segments):
        output = []
        for segment in segments:
            output.extend(segment)
            if segment:
                output.append(self.eos_token)
        if output:
            output.append(self.cls_token)
        return output

    def fix_alignment(self, segments):
        """Turn a nested segment alignment into an alignment for the whole input,
        by offsetting and accounting for special tokens."""
        offset = 0
        output = []
        for segment in segments:
            seen = set()
            for idx_group in segment:
                output.append([idx + offset for idx in idx_group])
                seen.update({idx for idx in idx_group})
            offset += len(seen)
            if segment:
                offset += 1
        return output


def clean_accents(text):
    return "".join(
        c for c in unicodedata.normalize("NFD", text) if unicodedata.category(c) != "Mn"
    )


def clean_fractions(text):
    chars = []
    for c in text:
        try:
            name = unicodedata.name(c)
        except ValueError:
            chars.append(c)
            continue
        name = unicodedata.name(c)
        if name.startswith("VULGAR FRACTION"):
            chars.append(unicodedata.normalize("NFKC", c))
        else:
            chars.append(c)
    return "".join(chars)


def clean_extended_unicode(text):
    return "".join(i for i in text if 31 < ord(i) < 127)


def serialize_bpe_ranks(data):
    return [{"key": list(key), "value": value} for key, value in data.items()]


def deserialize_bpe_ranks(data):
    return {tuple(item["key"]): item["value"] for item in data}



from spacy.pipeline import Pipe
from spacy.util import minibatch
import re
import numpy

from ..util import get_pytt_tokenizer, flatten_list, get_sents


class PyTT_WordPiecer(Pipe):
    """spaCy pipeline component to assign PyTorch-Transformers word-piece
    tokenization to the Doc, which can then be used by the token vector
    encoder. Note that this component doesn't modify spaCy's tokenization. It
    only sets extension attributes and aligns the tokens."""

    name = "pytt_wordpiecer"

    @classmethod
    def from_nlp(cls, nlp, **cfg):
        """Factory to add to Language.factories via entry point."""
        return cls(nlp.vocab, **cfg)

    @classmethod
    def from_pretrained(cls, vocab, pytt_name, **cfg):
        model = get_pytt_tokenizer(pytt_name).from_pretrained(pytt_name)
        return cls(vocab, model=model, pytt_name=pytt_name, **cfg)

    @classmethod
    def Model(cls, pytt_name, **kwargs):
        return get_pytt_tokenizer(pytt_name).blank()

    def __init__(self, vocab, model=True, **cfg):
        """Initialize the component.

        vocab (spacy.vocab.Vocab): The spaCy vocab to use.
        model: Not used here.
        **cfg: Optional config parameters.
        """
        self.vocab = vocab
        self.cfg = cfg
        self.model = model

    def __call__(self, doc):
        """Apply the pipe to one document. The document is
        modified in-place, and returned.

        Both __call__ and pipe should delegate to the `predict()`
        and `set_annotations()` methods.
        """
        self.require_model()
        pieces = self.predict([doc])
        self.set_annotations([doc], pieces)
        return doc

    @property
    def alignment_strategy(self):
        """Mostly used for debugging to make the WordPiecer raise error on
        misaligned tokens and prevent forcing alignment.
        """
        retry_alignment = self.cfg.get("retry_alignment", False)
        force_alignment = self.cfg.get("force_alignment", True)
        return (retry_alignment, force_alignment)

    def pipe(self, stream, batch_size=128):
        """Process Doc objects as a stream and assign the extracted features.

        stream (iterable): A stream of Doc objects.
        batch_size (int): The number of texts to buffer.
        YIELDS (spacy.tokens.Doc): Processed Docs in order.
        """
        for docs in minibatch(stream, size=batch_size):
            docs = list(docs)
            outputs = self.predict(docs)
            self.set_annotations(docs, outputs)
            for doc in docs:
                yield doc

    def predict(self, docs):
        """Run the word-piece tokenizer on a batch of docs and return the
        extracted strings.

        docs (iterable): A batch of Docs to process.
        RETURNS (tuple): A (strings, None) tuple.
        """
        output = []
        for doc in docs:
            doc_words = []
            doc_align = []
            offset = 0
            for sent in get_sents(doc):
                sent_words = []
                sent_align = []
                for segment in sent._.pytt_segments:
                    seg_words = self.model.tokenize(segment.text)
                    seg_words, seg_align = self._align(
                        segment, seg_words, offset=offset
                    )
                    assert len(segment) == len(seg_align)
                    sent_words.append(seg_words)
                    sent_align.append(seg_align)
                sw_flat = self.model.add_special_tokens(sent_words)
                sa_flat = self.model.fix_alignment(sent_align)
                doc_words.extend(sw_flat)
                doc_align.extend(sa_flat)
                offset += len(sw_flat)
            output.append((doc_words, doc_align))
        return output

    def _align(self, segment, wp_tokens, *, offset=0):
        retry, force = self.alignment_strategy
        spacy_tokens = [self.model.clean_token(w.text) for w in segment]
        new_wp_tokens = [self.model.clean_wp_token(t) for t in wp_tokens]
        assert len(wp_tokens) == len(new_wp_tokens)
        align = align_word_pieces(spacy_tokens, new_wp_tokens, retry=retry)
        if align is None and not force:
            spacy_string = "".join(spacy_tokens).lower()
            wp_string = "".join(new_wp_tokens).lower()
            print("spaCy:", spacy_string)
            print("WP:", wp_string)
            raise AssertionError((spacy_string, wp_string))
        elif align is None and force:
            # As a final fallback, we resort to word-piece tokenizing
            # the spaCy tokens individually, to make the alignment
            # trivial.
            wp_tokens, align = _tokenize_individual_tokens(self.model, segment)
        for indices in align:
            for i in range(len(indices)):
                indices[i] += offset
        return wp_tokens, align

    def set_annotations(self, docs, outputs, tensors=None):
        """Assign the extracted tokens and IDs to the Doc objects.

        docs (iterable): A batch of `Doc` objects.
        outputs (iterable): A batch of outputs.
        """
        # Set model.max_len to some high value, to avoid annoying prints.
        max_len = self.model.max_len
        self.model.max_len = 1e12
        for doc, (wordpieces, alignment) in zip(docs, outputs):
            doc._.pytt_word_pieces_ = wordpieces
            doc._.pytt_word_pieces = self.model.convert_tokens_to_ids(wordpieces)
            doc._.pytt_alignment = alignment
            nr_word = len(doc._.pytt_word_pieces)
            words_per_sent = sum(
                len(sent._.pytt_word_pieces) for sent in get_sents(doc)
            )
            if nr_word != words_per_sent:
                print([repr(w.text) for w in doc])
                for sent in get_sents(doc):
                    print(sent._.pytt_word_pieces_)
                    for w in sent:
                        print(w.text, w._.pytt_alignment)
                print(doc._.pytt_word_pieces_)
                raise ValueError(
                    f"Error calculating word pieces for sentences. Total number "
                    f"of wordpieces in the doc was {nr_word}, but adding up the "
                    f"wordpieces for its sentences we get {words_per_sent}. This "
                    f"means there's a bug in the extension attributes or "
                    f"the tokenizer.add_special_tokens() logic, often when "
                    f"a spaCy sentence aligns against 0 wordpieces."
                )
        self.model.max_len = max_len

    def use_params(self, params):
        yield


alpha_re = re.compile(r"[^A-Za-z]+")


def align_word_pieces(spacy_tokens, wp_tokens, retry=True):
    """Align tokens against word-piece tokens. The alignment is returned as a
    list of lists. If alignment[3] == [4, 5, 6], that means that spacy_tokens[3]
    aligns against 3 tokens: wp_tokens[4], wp_tokens[5] and wp_tokens[6].
    All spaCy tokens must align against at least one element of wp_tokens.
    """
    spacy_tokens = list(spacy_tokens)
    wp_tokens = list(wp_tokens)
    if not wp_tokens:
        return [[] for _ in spacy_tokens]
    elif not spacy_tokens:
        return []
    # Check alignment
    spacy_string = "".join(spacy_tokens).lower()
    wp_string = "".join(wp_tokens).lower()
    if not spacy_string and not wp_string:
        return None
    if spacy_string != wp_string:
        if retry:
            # Flag to control whether to apply a fallback strategy when we
            # don't align, of making more aggressive replacements. It's not
            # clear whether this will lead to better or worse results than the
            # ultimate fallback strategy, of calling the sub-tokenizer on the
            # spaCy tokens. Probably trying harder to get alignment is good:
            # the ultimate fallback actually *changes what wordpieces we
            # return*, so we get (potentially) different results out of the
            # transformer. The more aggressive alignment can only change how we
            # map those transformer features to tokens.
            spacy_tokens = [alpha_re.sub("", t) for t in spacy_tokens]
            wp_tokens = [alpha_re.sub("", t) for t in wp_tokens]
            spacy_string = "".join(spacy_tokens).lower()
            wp_string = "".join(wp_tokens).lower()
            if spacy_string == wp_string:
                return _align(spacy_tokens, wp_tokens)
        # If either we're not trying the fallback alignment, or the fallback
        # fails, we return None. This tells the wordpiecer to align by
        # calling the sub-tokenizer on the spaCy tokens.
        return None
    output = _align(spacy_tokens, wp_tokens)
    if len(set(flatten_list(output))) != len(wp_tokens):
        return None
    return output


def _align(seq1, seq2):
    # Map character positions to tokens
    map1 = _get_char_map(seq1)
    map2 = _get_char_map(seq2)
    # For each token in seq1, get the set of tokens in seq2
    # that share at least one character with that token.
    alignment = [set() for _ in seq1]
    unaligned = set(range(len(seq2)))
    for char_position in range(map1.shape[0]):
        i = map1[char_position]
        j = map2[char_position]
        alignment[i].add(j)
        if j in unaligned:
            unaligned.remove(j)
    # Sort, make list
    output = [sorted(list(s)) for s in alignment]
    # Expand alignment to adjacent unaligned tokens of seq2
    for indices in output:
        if indices:
            while indices[0] >= 1 and indices[0] - 1 in unaligned:
                indices.insert(0, indices[0] - 1)
            last = len(seq2) - 1
            while indices[-1] < last and indices[-1] + 1 in unaligned:
                indices.append(indices[-1] + 1)
    return output


def _get_char_map(seq):
    char_map = numpy.zeros((sum(len(token) for token in seq),), dtype="i")
    offset = 0
    for i, token in enumerate(seq):
        for j in range(len(token)):
            char_map[offset + j] = i
        offset += len(token)
    return char_map


def _tokenize_individual_tokens(model, sent):
    # As a last-chance strategy, run the wordpiece tokenizer on the
    # individual tokens, so that alignment is trivial.
    wp_tokens = []
    sent_align = []
    offset = 0
    for token in sent:
        if token.text.strip():
            subtokens = model.tokenize(token.text)
            wp_tokens.extend(subtokens)
            sent_align.append([i + offset for i in range(len(subtokens))])
            offset += len(subtokens)
        else:
            sent_align.append([])
    return wp_tokens, sent_align



import pytest
from numpy.testing import assert_equal
from spacy_pytorch_transformers import PyTT_TokenVectorEncoder
from spacy.vocab import Vocab
import pickle

from .util import make_tempdir, is_valid_tensor


@pytest.fixture
def docs(nlp):
    texts = ["the cat sat on the mat.", "hello world."]
    return [nlp(text) for text in texts]


@pytest.fixture(scope="session")
def tok2vec(name, nlp):
    return nlp.get_pipe("pytt_tok2vec")


def test_from_pretrained(tok2vec, docs):
    docs_out = list(tok2vec.pipe(docs))
    assert len(docs_out) == len(docs)
    for doc in docs_out:
        diff = doc.tensor.sum() - doc._.pytt_last_hidden_state.sum()
        assert abs(diff) <= 1e-2


def test_set_annotations(name, tok2vec, docs):
    scores = tok2vec.predict(docs)
    tok2vec.set_annotations(docs, scores)
    for doc in docs:
        assert doc._.pytt_last_hidden_state is not None
        if "bert" in name:
            assert doc._.pytt_pooler_output is not None
        assert doc._.pytt_d_last_hidden_state is not None
        assert doc._.pytt_d_last_hidden_state.ndim == 2
        assert doc._.pytt_d_pooler_output.ndim == 2


def test_begin_finish_update(tok2vec, docs):
    scores, finish_update = tok2vec.begin_update(docs)
    finish_update(docs)


@pytest.mark.parametrize(
    "text1,text2,is_similar,threshold",
    [
        ("The dog barked.", "The puppy barked.", True, 0.5),
        ("rats are cute", "cats please me", True, 0.6),
    ],
)
def test_similarity(nlp, text1, text2, is_similar, threshold):
    doc1 = nlp(text1)
    doc2 = nlp(text2)
    similarity = doc1.similarity(doc2)
    if is_similar:
        assert similarity >= threshold
    else:
        assert similarity < threshold


def test_tok2vec_to_from_bytes(tok2vec, docs):
    doc = tok2vec(docs[0])
    assert is_valid_tensor(doc.tensor)
    bytes_data = tok2vec.to_bytes()
    new_tok2vec = PyTT_TokenVectorEncoder(Vocab(), **tok2vec.cfg)
    with pytest.raises(ValueError):
        new_doc = new_tok2vec(docs[0])
    new_tok2vec.from_bytes(bytes_data)
    new_doc = new_tok2vec(docs[0])
    assert is_valid_tensor(new_doc.tensor)
    assert_equal(doc.tensor, new_doc.tensor)


def test_tok2vec_to_from_disk(tok2vec, docs):
    doc = tok2vec(docs[0])
    assert is_valid_tensor(doc.tensor)
    with make_tempdir() as tempdir:
        file_path = tempdir / "tok2vec"
        tok2vec.to_disk(file_path)
        new_tok2vec = PyTT_TokenVectorEncoder(Vocab())
        new_tok2vec.from_disk(file_path)
    new_doc = new_tok2vec(docs[0])
    assert is_valid_tensor(new_doc.tensor)
    assert_equal(doc.tensor, new_doc.tensor)


# Fails on 3.6 due to type annotations :(. See
# https://github.com/cloudpipe/cloudpickle/pull/299
@pytest.mark.xfail
def test_tok2vec_pickle_dumps_loads(tok2vec, docs):
    doc = tok2vec(docs[0])
    assert is_valid_tensor(doc.tensor)
    pkl_data = pickle.dumps(tok2vec)
    new_tok2vec = pickle.loads(pkl_data)
    new_doc = new_tok2vec(docs[0])
    assert is_valid_tensor(new_doc.tensor)
    assert_equal(doc.tensor, new_doc.tensor)
